##
## Author Michel Sanner Feb 2009
##
## Copyrights TSRI and M. Sanner
##

from Scenario2.events import AddMAAEvent, RemoveMAAEvent, CurrentFrameEvent,\
     PlayerStartEvent, PlayerStopEvent
from Scenario2.sequenceAnimator import SequenceAnimator

import Tkinter, Pmw, os
from mglutil.util.callback import CallBackFunction
from mglutil.util.packageFilePath import findFilePath
from mglutil.gui.BasicWidgets.Tk.thumbwheel import ThumbWheel
from Scenario2.playerControls import PlayerControls

class SequenceAnimatorGUI:
    """
    class displaying the list of MultipleActorActions stored in
    a clipboard object and allowing to play them in sequence and to rearrange
    their order.
    """

    def __init__(self, animator, master=None):
        """
        SequenceAnimatorGUI constructor
        
        SequenceAnimatorGUI <- SequenceAnimatorGUI(animator, master=None)

        animator is an instance of a SequenceAnimator object
        master is a Tk container object. If master is None a Tkinter.Toplevel
        object is created

        This object registers interest in AddMAAToClipBoardEvent and
        RemoveMAAFromClipBoardEvent events with the standard
        SequenceAnimator object
        """

        assert isinstance(animator, SequenceAnimator)
        
        self.animator = animator
        
        self.master = master

        self.pl = None # will be the PlayerControls object
        # mpg video frame rate (based on mplayer info)
        self.fps = 24.0
        animator.registerListener(AddMAAEvent, self.addMaaEvent_cb)
        animator.registerListener(RemoveMAAEvent, self.removeMaaEvent_cb)
        animator.registerListener(CurrentFrameEvent, self.curFrameEvent_cb)
        animator.registerListener(PlayerStartEvent, self.playerStart_cb)
        animator.registerListener(PlayerStopEvent, self.playerStop_cb)
        self.createGUI()
        self.refreshGUI()


    def playerStart_cb(self, event=None):
        """This function is called when a PlayerStartEvent is created
        """
        # save current active MAA
        sel = self.MAAContainer.getcurselection()
        if len(sel):
            self._currentSel = self.maaList.index(sel[0])


    def playerStop_cb(self, event=None):
        """This function is called when a PlayerStopEvent is created
        """
        # restore current active MAA
        if hasattr(self, '_currentSel'):
            self.MAAContainer.select_set(self._currentSel)
            del self._currentSel
        

    def addMaaEvent_cb(self, event=None):
        """This function is called by the dispatchEvent() in addMAAat() method of MAADirector.
        It updates the MAA Chooser and the Player"""
        # update list of MAAs in list
        self.refreshGUI()
        # update frame counter in player controls obejct
        self.pl.update()
        

    def removeMaaEvent_cb(self, event=None):
        """This function is called by the dispatchEvent() in removeMAA() method of SequenceAnimator.
        It updates the MAA Chooser and the Player."""
        self.refreshGUI()
        self.selectionCommand_cb()
        self.pl.update()


    def curFrameEvent_cb(self, event=None):
        """
        call back for current frame change in sequence animator
        highlight the MAA contianing this frame
        """
        frame = event.frame
        lb = self.MAAContainer
        for n, val in enumerate(self.animator.maas):
            maa, firstFrame = val
            if frame<=firstFrame or frame>firstFrame+maa.lastPosition:
                # frame not in this maa
                lb.itemconfigure(n, selectbackground='') # default grey
                lb.select_clear(n)
            else:
                lb.itemconfigure(n, selectbackground='yellow') # default grey
                lb.select_set(n)
        self.pl.update()

                
    def createGUI(self):
        """
        Create a ScrolledFrame to hold MAA entries
        """
        if self.master is None:
            self.master = master = Tkinter.Toplevel()
            self.ownsMaster = True
        else:
            self.ownsMaster = False
        self.balloon = Pmw.Balloon(self.master)

        ##
        ## Action Container
        ##        
        parent =  Pmw.Group(self.master, tag_text='Actions')

        # create container
	w = self.MAAContainer = Pmw.ScrolledListBox(
            parent.interior(), #labelpos='nw', label_text='AnimationClips',
            listbox_height = 4,
            selectioncommand = self.selectionCommand_cb,
            #dblclickcommand=self.defCmd,
            usehullsize = 0,
            #hull_width = 300,
            #hull_height = 200,
            )
        ##
        ## Action manipulation group
        ##        
	#w.grid(sticky='nesw', column=0, row=0, rowspan=3)
        w.pack(fill='both', expand=1, side="left")
        w._listbox.config(exportselection = False)
        
        ICONDIR = findFilePath('icons', 'mglutil.gui.BasicWidgets.Tk')
        manipGroup =  Pmw.Group(parent.interior(),
                                 tag_text='Manipulate Actions')
        master = manipGroup.interior()
        fr1 = Tkinter.Frame(master)
        fr1.pack(fill='both', expand=1, side='top')
        self.gotoStartIcon = im = Tkinter.PhotoImage(
            file=os.path.join(ICONDIR, 'go_to_start.gif'), master=fr1)
        w = self.startActionb =  Tkinter.Button(
            fr1, image=im, command=self.gotoActionStart_cb)
        w.pack(side='left', anchor='n')
        self.balloon.bind(w, "goto action's start") 

        self.gotoEndIcon = im = Tkinter.PhotoImage(
            file=os.path.join(ICONDIR, 'go_to_end.gif'), master=fr1)
        w = self.endActionb =  Tkinter.Button(
            fr1, image=im, command=self.gotoActionEnd_cb)
        w.pack(side='left', anchor='n')
        self.balloon.bind(w, "goto action's end") 
        
        self.moveUpIcon = im = Tkinter.PhotoImage(
            #file=os.path.join(ICONDIR, 'arrowup.gif'), master=master)
            file=os.path.join(ICONDIR, 'arrow_up.gif'), master=fr1)
        w = self.upb =  Tkinter.Button(fr1, image=im, command=self.moveUp_cb)
        w.pack(side='left', anchor='n')
        self.balloon.bind(w, "Move selected action one position up") 
        
        self.moveDownIcon = im = Tkinter.PhotoImage(
            #file=os.path.join(ICONDIR, 'arrowdown.gif'), master=master)
            file=os.path.join(ICONDIR, 'arrow_down.gif'), master=fr1)
        w = self.downb =  Tkinter.Button(
            fr1, image=im, command=self.moveDown_cb)
        w.pack(side='left', anchor='n')
        self.balloon.bind(w, "Move selected action one position down")         
        
        #self.deleteIcon = im = Tkinter.PhotoImage(
        #    file=os.path.join(ICONDIR, 'close.gif'), master=master)
        #w = self.deleteb =  Tkinter.Button(
        #    master, image=im, command=self.delete_cb)
        fr2 = Tkinter.Frame(master)
        fr2.pack(fill='both', expand=1, side='top')
        w = self.deleteb =  Tkinter.Button(
          fr2, text='delete', command=self.delete_cb)
        w.pack(side='top', anchor='nw')
        self.balloon.bind(w, "Delete selected action")         

        manipGroup.pack(side='top', fill='x')

        ##
        ## Action Group
        ##
        self.actionGroup =  Pmw.Group(parent.interior(), tag_text='force')
        actionFrame = self.actionGroup.interior()
        self.forceOrientTk = Tkinter.IntVar()
        w = self.orientb =  Tkinter.Checkbutton(
            actionFrame, text='Orientation', variable=self.forceOrientTk,
            command=self.orient_cb)
        w.pack(side='top', fill='x', anchor='nw')
        w.configure(state='disabled')
        self.balloon.bind(w, "Force the original orientation on the first frame of this action")
        
        self.forceRenderingTk = Tkinter.IntVar()
        w = self.renderingb =  Tkinter.Checkbutton(
            actionFrame, text='Rendering', variable=self.forceRenderingTk,
            command=self.rendering_cb)
        w.pack(side='top', fill='x', anchor='nw')
        w.configure(state='disabled')
        self.balloon.bind(w, "Force the original rendering on the first frame of this action")    
        self.actionGroup.pack(side='top', anchor='n', fill='x')

        w = self.editb =  Tkinter.Button(parent.interior(), text='Edit ...',
                                         command=self.edit_cb)
        w.pack(side='top', fill='x', anchor='n')
        self.balloon.bind(w, "Edit selected action")    


        ##
        ## Start Group
        ##
        startgroup =  Pmw.Group(parent.interior(), tag_text='Start')
        startframe = startgroup.interior()
        self.startW = sw = Pmw.RadioSelect(
            startframe, selectmode='single', orient='vertical',
            buttontype='radiobutton', command=self.setStartFlag_cb)

        for text in ['after previous', 'with previous']:
            sw.add(text)

        sw.pack(side='top', anchor='n', fill='x') # padx=5, pady=5)

        self.balloon.bind(sw.button("after previous"),
                          "Start selected item after the previous one")
        self.balloon.bind(sw.button("with previous"),
                          "Start selected item with the previous one")

        #sw.setvalue('after previous')
        self.offsetw = offw = Pmw.Counter(
            startframe, labelpos = 'w', label_text = 'Offset:',
            entry_width = 6, entryfield_value = 0,
            entryfield_command = self.setOffset_cb,
            datatype = self.setOffset_cb,
            entryfield_validate = {'validator' : 'integer' } )

        self.balloon.bind(offw,
                          "positive or negative offset from starting position")

        offw.pack(side='top', anchor='n', fill='x')
        parent.pack(fill='both', expand=1, padx=5, pady=5)
        startgroup.pack(side='top')
        
        # add a thumb wheel for setting video (mpg) duration
        ## self.mpgDuration = ThumbWheel(parent.interior(),
        ##                    showLabel=1, width=60, height=16,
        ##                    type=int, value=0,
        ##                    callback=self.setMpgDuration_cb,
        ##                    continuous=False, oneTurn=10,
        ##                    min=0, wheelPad=2,
        ##                    labCfg = {'text':'Video(mpg) duration(sec):',
        ##                               'side':'top'})
        # add an EntryField for setting video (mpg) duration
        self.mpgDuration =Pmw.EntryField(parent.interior(),
                labelpos = 'n',
                label_text = 'fps:24, duration(sec)',
                value = 0, command=self.setMpgDuration_cb,
                validate = {'validator':'real', 'min':0},
                                         entry_width=10
                )
 
        self.mpgDuration.pack(side='top', anchor='w', pady=5, expand=1)# fill='x')
        

        #self.frameCounter = Tkinter.Label(master, text='Frame 000000')
        #self.frameCounter.pack(side='top')
        frame = Tkinter.Frame(self.master)               
        self.pl = PlayerControls( self.animator, root=frame, #root=self.master,
                                  form2=1, hasSlider=True)
        frame.pack(side='top')


        self.selectionCommand_cb() # called to configure the buttons


    def gotoActionEnd_cb(self, event=None):
        """
        callback for got to end button
        """
        curSel = self.MAAContainer.getcurselection()
        if len(curSel)==0:
            return
        index = self.maaList.index(curSel[0])
        maa, pos = self.animator.maas[index]
        #print 'Going to frame', pos + maa.lastPosition
        director = self.pl.director
        director.setValuesAt(pos + maa.lastPosition)
        director.viewerRedraw()
        self.MAAContainer.select_set(index)


    def gotoActionStart_cb(self, event=None):
        """
        callback for got to start button
        """
        curSel = self.MAAContainer.getcurselection()
        if len(curSel)==0:
            return
        index = self.maaList.index(curSel[0])
        maa, pos = self.animator.maas[index]
        #print 'Going to frame', pos
        director = self.pl.director
        director.setValuesAt(pos)
        director.viewerRedraw()
        self.MAAContainer.select_set(index)
        
        
    def refreshGUI(self, updateMpg=True):
        """Updates the entries of the MAA Chooser."""
        #self.maaList = ["%05d: "%i+ a.name for a,i in self.animator.maas]

        # save current selection
        curSel = self.MAAContainer.getcurselection()
        if len(curSel):
            indexSel = self.maaList.index(curSel[0])
        
        self.maaList = maas = []
        for i, (a,pos) in enumerate(self.animator.maas):
            if a.startFlag == 'with previous' and i != 0:
                maas.append("    %05d-%05d: "%(pos, pos+a.lastPosition) + a.name)
            else:
                maas.append("%05d-%05d: "%(pos, pos+a.lastPosition)+ a.name)

        self.MAAContainer.setlist(self.maaList)
        if len(curSel):
            self.MAAContainer.select_set(indexSel)
        if updateMpg:
            self.mpgDuration.setvalue(str(round(self.getMpgDuration(), 2)))

            
    def moveUp_cb(self, event=None):
        """Callback of the 'Move Up' button. Moves selected MAA entry up one position in the
        MAA Chooser."""
        sel = self.MAAContainer.getcurselection()
        if not len(sel): return
        index = self.maaList.index(sel[0])
        maa = self.animator.maas[index]
        self.animator.moveMAA(maa, index-1)
        self.refreshGUI()
        #self.MAAContainer.setvalue("%05d: %s"%(maa[1],maa[0].name))
        self.MAAContainer.setvalue(self.maaList[index-1])
        self.selectionCommand_cb()

        
    def moveDown_cb(self, event=None):
        """Callback of the 'Move Down' button. Moves selected MAA entry down one position in the
        MAA Chooser."""
        sel = self.MAAContainer.getcurselection()
        if not len(sel):return
        index = self.maaList.index(sel[0])
        maa = self.animator.maas[index]
        self.animator.moveMAA(maa, index+1)
        self.refreshGUI()
        #self.MAAContainer.setvalue("%05d: %s"%(maa[1],maa[0].name))
        self.MAAContainer.setvalue(self.maaList[index+1])
        self.selectionCommand_cb()


    def delete_cb(self, event=None):
        """Callback of the 'Delete' button. Deletes selected MAA from the MAA Chooser and the
        Sequence Animator's maas list.
        """
        sel = self.MAAContainer.getcurselection()
        if len(sel) == 0:
            return
        index = self.maaList.index(sel[0])
        maa, position = self.animator.maas[index]
        self.animator.removeMAA(maa, position)
        
        
    def orient_cb(self, event=None):
        """Callback of the 'Orinetation' button.
        """
        sel = self.MAAContainer.getcurselection()
        if len(sel)==0:
            return
        index = self.maaList.index(sel[0])
        maa, position = self.animator.maas[index]
        if hasattr(maa, "interpolateOrient"):
            maa.interpolateOrient = self.forceOrientTk.get()
        else:
            maa.forceOrient = self.forceOrientTk.get()
        
        
    def rendering_cb(self, event=None):
        """Callback of the 'Rendering' button.
        """
        sel = self.MAAContainer.getcurselection()
        if len(sel)==0:
            return
        index = self.maaList.index(sel[0])
        maa, position = self.animator.maas[index]
        if hasattr(maa, "interpolateRendering"):
            maa.interpolateRendering = self.forceRenderingTk.get()
        else:
            maa.forceRendering = self.forceRenderingTk.get()
        
 
    def sort_cb(self, tag):
        """Callback of the 'Sort Poly.' radio button.
        """
        curSel = self.MAAContainer.getcurselection()
        if len(curSel)==0:
            return
        index = self.maaList.index(curSel[0])
        maa, position = self.animator.maas[index]
        maa.sortPoly = tag

        
    def edit_cb(self, event=None):
        """Callback of the 'Edit ...' button.
        displays MAA editor for the currently slected action
        """
        sel = self.MAAContainer.getcurselection()
        if len(sel) == 0:return
        index = self.maaList.index(sel[0])
        maa, position = self.animator.maas[index]

        kw = maa.editorKw
        args = maa.editorArgs
        kw['master'] = self.master
        editor =  maa.editorClass( *args, **kw)
        editor.edit(maa)
        # check if the maa has been added more than once to the sequence animator:
        for i , _maa in enumerate(self.animator.maas):
            if _maa[0] == maa and i < index:
                index = i
                position = _maa[1]
                break
        self.update(index, position)
        

    def update(self, index, position, updateMpg=True):
        self.animator.updatePositions(index, position)
        self.refreshGUI(updateMpg=updateMpg)
        self.pl.update()
        
        
    def selectionCommand_cb(self, event=None):
        """This function is called when an MAA is selected in the MAA Chooser.
        """
        sel = self.MAAContainer.getcurselection()
        self.upb.configure(state='disabled')   # move up button
        self.downb.configure(state='disabled') # move down button
        self.deleteb.configure(state='disabled') # delete button
            
        if len(sel)==0:
            return

        self.deleteb.configure(state='normal')

        index = self.maaList.index(sel[0])
        
        if index > 0:
            self.upb.configure(state='normal')

        if index < len(self.maaList)-1:
            self.downb.configure(state='normal')
        maa, pos = self.animator.maas[index]

        if hasattr(maa, "interpolateOrient"):
            self.orientb.configure(state='normal')
            self.forceOrientTk.set(maa.interpolateOrient)
            self.actionGroup._tag.configure(text = 'interpolate')
            self.balloon.bind(self.orientb, "Interpolate orientation during playback")
        elif hasattr(maa, 'forceOrient'):
            self.orientb.configure(state='normal')
            self.forceOrientTk.set(maa.forceOrient)
            self.actionGroup._tag.configure(text = 'force')
            self.balloon.bind(self.orientb, "Force the original orientation on the first frame of this action")
        else:
            self.orientb.configure(state='disabled')
            
        if hasattr(maa, "interpolateRendering"):
            self.renderingb.configure(state='normal')
            self.forceRenderingTk.set(maa.interpolateRendering)
            self.balloon.bind(self.renderingb, "Interpolate rendering during playback")
        elif hasattr(maa, 'forceRendering'):
            self.renderingb.configure(state='normal')
            self.forceRenderingTk.set(maa.forceRendering)
            self.balloon.bind(self.renderingb, "Force the original rendering on the first frame of this action")
        else:
            self.renderingb.configure(state='disabled')
            
        if maa.editorClass:
            self.editb.configure(state='normal')
        else:
            self.editb.configure(state='disabled')
        
           
        self.startW.setvalue(maa.startFlag)
        self.offsetw.setvalue(maa.startOffset)
        startWState = 'normal'
        if index == 0:
            startWState = 'disabled'
        for i in range(2):
            self.startW.button(i).configure(state=startWState)
        

    def setStartFlag_cb(self, startFlag = None):
        """Callback of the 'Start' radio buttons. Sets a new starting
        position of the selected maa."""
        #print "setStartFlag_cb", startFlag
        sel = self.MAAContainer.getcurselection()
        if not len(sel):return
        index = self.maaList.index(sel[0])
        maa, position = self.animator.maas[index]
        position, index  = self.animator.computePosition(maa, startFlag=startFlag)
        maa.startFlag = startFlag

        self.animator.updatePositions(index, position-maa.startOffset)
        self.refreshGUI()
        #self.MAAContainer.setvalue(self.maaList[index])
        # make sure the player end frame is correct
        self.pl.update()

                
    def setOffset_cb(self, text=None, factor=None, increment=None, **kw):
        # text is current content of entry
        # factor is 1 for increment and -1 for decrement
        # increment is value of increment megawidget option
        """Callback of the 'offset' entry field.
        Sets maa.startOffset for the currently selected action

        Has to return the coutner value
        """
        sel = self.MAAContainer.getcurselection()
        if not len(sel):
            return 0
        
        if text is not None: # incrment or decrement arrow was used
            value = int(text)+(factor*increment)
        else: # return was hit in entry
            value = int(self.offsetw.get())
            
        index = self.maaList.index(sel[0])
        maa, oldpos = self.animator.maas[index]

        # remove offset form the current position
        position = oldpos - maa.startOffset

        maa.startOffset = value

        self.animator.updatePositions(index, position)

        self.refreshGUI()

        # make sure the player end frame is correct
        self.pl.update()

        return value


    def getMpgDuration(self):
        self.animator.updateEndFrame()
        nframes = self.animator.endFrame
        #print "n frames:", nframes
        return float(nframes)/self.fps

        
    def setMpgDuration_cb(self):
        if len(self.animator.maas) == 0:
            return
        duration = float(self.mpgDuration.getvalue())
        curduration=self.getMpgDuration()
        ratio = duration/curduration
        nfr = 0
        for item in self.animator.maas:
            maa = item[0]
            fpos =  maa.firstPosition
            lpos =  maa.lastPosition
            nframes = (lpos - fpos)+1
            nfr += nframes
            nframes1 = int(round(nframes*ratio))
            if nframes1<1: nframes1=1
            kw = {'kfpos': [fpos, fpos+nframes1]}
            try:
                maa.configure(kfpos=[fpos, fpos+nframes1])
            except TypeError:
                maa.configure(nbFrames=nframes1)
                
            maa.startOffset = int(round(maa.startOffset * ratio))
            
        self.update(0, self.animator.maas[0][1],updateMpg=False)
        duration = self.getMpgDuration()
        self.mpgDuration.setvalue(str(round(duration, 2)))
        #print "duration:",  duration



from Scenario2.keyframes import ValueObject, ValueGenerator,KF, Interval, \
     AutoCurrentValueObject, KFAutoCurrentValue
from Scenario2.interpolators import FloatScalarInterpolator
from Scenario2.actions import Actions
from Scenario2.actor import CustomActor


def test_001constructors():
    a = Actions()


def test_002addKeyframe1():
    """
    Test adding KFs that fall on nothing or an existing KF.
    If a new KF falls on to existing one, the value of the existing Kf is
    overwritten with the value of the new KF.
    A linked KF cannot be added on top of an existing KF
    """
    a = Actions()
    
    # add keyframe as position value pair
    # dest. : 0
    a.addKeyframe( (0, 10) )
    assert len(a.keyframes) == 1
    assert a.keyframes[0].getValue() == 10
    kf0 = a.keyframes[0]
    
    # add keyframe object
    # dest. : 0        10
    kf1 = KF( 10, 25 )
    a.addKeyframe( kf1 )
    assert len(a.keyframes) == 2
    assert a.keyframes[1].getValue() == 25

    # add keyframe between the 2 existing ones
    # dest. : 0   5   10
    kf2 = KF( 5, 28 )
    a.addKeyframe( kf2 )
    assert len(a.keyframes) == 3
    assert a.keyframes[0] == kf0
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf1
    
    # overwrite existing keyframe value
    # dest. : 0   5   10
    kf3 = KF( 5, -12 )
    a.addKeyframe( kf3 )
    # no new KF is added
    assert len(a.keyframes) == 3
    assert a.keyframes[0] == kf0
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf1
    # value of kf2 is now -12
    assert a.keyframes[1].getValue() == -12

    # try to add a linked KF (this is not allowed): 
    kf4 = kf1.linkedCopy(5)
    assert a.addKeyframe(kf4) == None

    


def test_002addKeyframe2():
    """ Tests splitting an interval with a new keyframe
    """
    a = Actions()

    # add first interval from 10-30
    kf1 = KF( 10, 25 )
    kf2 = KF( 30, 35 )
    valGen = ValueGenerator()
    i1 = Interval(kf1, kf2, valGen)
    from Scenario2.interpolators import IntScalarInterpolator
    ip = IntScalarInterpolator(None, None)
    i1.setValGenGenerator(ip)
    a.addInterval(i1)
    #add a new keyframe that will split the first interval
    kf3 = KF(15, 45)
    a.addKeyframe( kf3 )
    kfs = a.keyframes
    assert kfs[0].pos == 10
    assert kfs[1].pos == 15
    assert kfs[2].pos == 30
    assert len(a.intervals) == 2
    
    in1 = a.intervals[0]
    assert in1.kf1 == kf1
    assert in1.kf2 == kf3
    vg1 = in1.valGen.generator
    assert isinstance(vg1, IntScalarInterpolator) == True
    assert vg1.firstVal == 25
    assert vg1.lastVal == 45


    in2 = a.intervals[1]
    assert in2.kf1 == kf3
    assert in2.kf2 == kf2
    vg2 = in2.valGen.generator
    assert isinstance(vg2, IntScalarInterpolator) == True
    assert vg2.firstVal == 45
    assert vg2.lastVal == 35

    assert kfs[0].leftInterval == None
    assert kfs[0].rightInterval == in1
    assert kfs[1].leftInterval == in1
    assert kfs[1].rightInterval == in2
    assert kfs[2].leftInterval == in2
    assert kfs[2].rightInterval == None
        
    # now there are 3 keyframes and 2 intervals.
    # create a KF that is linked copy of kf1  and add it to the second interval
    kf4 = kf1.linkedCopy(20)
    a.addKeyframe( kf4 )
    kfs = a.keyframes
    #for i, kf in enumerate(kfs):
    #    print "kf(%d)" %i, kf.pos, kf.getValue()
    assert kfs[0].pos == 10
    assert kfs[1].pos == 15
    assert kfs[2].pos == 20
    assert kfs[3].pos == 30
    assert len(a.intervals) == 3

    assert kfs[0].isLinked() == True
    assert kfs[2].isLinked() == True

    in1 = a.intervals[1]
    assert in1.kf1 == kf3
    assert in1.kf2 == kf4
    vg1 = in1.valGen.generator
    assert vg1.firstVal == 45
    assert vg1.lastVal == 25

    in2 = a.intervals[2]
    assert in2.kf1 == kf4
    assert in2.kf2 == kf2
    vg2 = in2.valGen.generator
    assert vg2.firstVal == 25
    assert vg2.lastVal == 35

    

class Object:

    def __init__(self, val):
        self.value = val
        
    def get(self):
        return self.value

    def set(self, value):
        self.value = value
        

def test_002addKFAutoCurrentValue1():
    """ Test adding a special case keyframe (KFAutoCurrentValue) whose value is generated by
    AutoCurrentValueObject. 
    """
    # test creating an action with a KFAutoCurrentValue kf  and a KF
    a = Actions()
    obj = Object(0)
    from Scenario2.actor import CustomActor
    from Scenario2.interpolators import IntScalarInterpolator
    actor = CustomActor('obj', obj, setMethod='set', getFunction=(obj.get, (), {}),
                        interp = (IntScalarInterpolator, (None, None), {} ))
    autoval = AutoCurrentValueObject(actor)
    kf1 = KFAutoCurrentValue(0, autoval)
    kf2 = KF (10, 20)
    i1 = Interval(kf1, kf2, generator = actor.behaviorList)
    a.addInterval(i1)

    assert a.getValue(0) == 0
    assert a.getValue(10) == 20
    assert a.getValue(5) == 10

    # change the value of the object
    obj.set(10)

    assert a.getValue(0) == 10  # the first keyframe updated properly
    assert a.getValue(10) == 20
    assert a.getValue(5) == 15



def test_002addKFAutoCurrentValue2():
    """
    Test replacing an existing keyframe with a KFAutoCurrentValue keyframe
    """
    # create an action with 3 keyframes and 2 intervals

    from Scenario2.actor import CustomActor
    from Scenario2.interpolators import IntScalarInterpolator
    obj = Object(10)
    actor = CustomActor('obj', obj, setMethod='set', getFunction=(obj.get, (), {}),
                        interp = (IntScalarInterpolator, (None, None), {} ))
    a = Actions(actor)
    kf1 = KF(0, 0)
    kf2 = KF(10, 20)
    kf3 = KF(10, 20)
    kf4 = KF(20, 30)
    gen1 =  actor.behaviorList
    gen2 = IntScalarInterpolator(None, None)
    i1 = Interval(kf1, kf2, generator = gen1)
    i2 = Interval(kf3, kf4, generator = gen2)
    a.addInterval(i1)
    a.addInterval(i2)
    
    assert a.getValue(0) == 0
    assert a.getValue(10) == 20
    assert a.getValue(5) == 10
    assert a.getValue(20) == 30
    assert a.getValue(15) == 25

    # create and add a new KFAutoCurrentValue keyframe at the position of the action's second keyframe
    autoval = AutoCurrentValueObject(actor)
    kf5 = KFAutoCurrentValue(10, autoval)
    
    a.addKeyframe(kf5)
    kfs = a.keyframes
    intervals = a.intervals
    assert len(kfs) == 3
    assert len(intervals) == 2
    assert kfs[1] == kf5    

    assert kfs[0].leftInterval == None
    assert kfs[0].rightInterval == intervals[0]
    assert intervals[0].kf1 == kfs[0]
    assert intervals[0].kf2 == kfs[1]
    assert kfs[1].leftInterval == intervals[0]
    assert kfs[1].rightInterval == intervals[1]
    assert intervals[1].kf1 == kfs[1]
    assert intervals[1].kf2 == kfs[2]
    assert kfs[2].leftInterval == intervals[1]
    assert kfs[2].rightInterval == None
    assert intervals[0].valGen.generator == gen1
    assert intervals[1].valGen.generator == gen2
    
    assert a.getValue(0) == 0
    #print  a.getValue(5), a.getValue(10), a.getValue(15), a.getValue(20), 
    assert a.getValue(5) == 5
    assert a.getValue(10) == 10
    assert a.getValue(15) == 20
    assert a.getValue(20) == 30



    
def test_003addInterval1():
    """
    Test adding Intervals that fall on nothing
    """
    a = Actions()

    # add first interval from 10-20
    #           10----------20
    kf1 = KF( 10, 25 )
    kf2 = KF( 20, 35 )
    i1 = Interval(kf1, kf2)
    val = a.addInterval( i1 )
    assert val == True
    assert len(a.intervals) == 1
    assert a.intervals[0] == i1
    assert len(a.keyframes) == 2
    
    # add second interval from 50-60
    #           10----------20                               50--------60
    kf1 = KF( 50, 25 )
    kf2 = KF( 60, 35 )
    i2 = Interval(kf1, kf2)
    val = a.addInterval( i2 )
    assert val == True
    assert len(a.intervals) == 2
    a.intervals[0] = i1
    a.intervals[1] = i2
    assert len(a.keyframes) == 4

    # add third interval from 30-40
    #           10----------20        30--------40           50--------60
    kf1 = KF( 30, 25 )
    kf2 = KF( 40, 35 )
    i3 = Interval(kf1, kf2)
    val = a.addInterval( i3 )
    assert val == True
    assert len(a.intervals) == 3
    a.intervals[0] = i1
    a.intervals[1] = i3
    a.intervals[2] = i2
    assert len(a.keyframes) == 6
    kfs = a.keyframes
    assert kfs[0].pos == 10
    assert kfs[1].pos == 20
    assert kfs[2].pos == 30
    assert kfs[3].pos == 40
    assert kfs[4].pos == 50
    assert kfs[5].pos == 60


def test_004addInterval1():
    """
    Test adding Intervals that fall on existng keyframes
    """

    a = Actions()
    # dest. :  0
    kf0 = KF(0, 0.)
    a.addKeyframe( kf0 )

    # add first interval from 0-10
    # The value of kf1 overwrites the value of kf0
    # the interval gets added
    # source:  0----------10
    # dest. :  0
    kf1 = KF( 0, 25. )
    kf2 = KF( 10, 35. )
    i1 = Interval(kf1, kf2)
    val = a.addInterval( i1 )
    assert val == True
    assert len(a.intervals) == 1
    a.intervals[0] = i1
    assert len(a.keyframes) == 2
    assert a.keyframes[0]== kf0
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1]== kf2

    # add second interval from 10-20
    # The value of kf1 overwrites the value of kf0
    # the interval gets added
    # dest. :  0----------10
    # source:             10----------20
    kf3 = KF( 10, 30. )
    kf4 = KF( 20, 35. )
    i2 = Interval(kf3, kf4)
    val = a.addInterval( i2 )
    kfs = a.keyframes
    intervals = a.intervals
    assert val == True
    assert len(intervals) == 2
    a.intervals[0] = i1
    a.intervals[1] = i2
    assert len(kfs) == 3
    assert kfs[0] == kf0
    assert kfs[1] == kf2
    assert kfs[1].getValue() == kf3.getValue()
    assert kfs[2] == kf4

    assert intervals[0].kf1 == kfs[0]
    assert intervals[0].kf2 == kfs[1]
    assert intervals[1].kf1 == kfs[1]
    assert intervals[1].kf2 == kfs[2]


def test_005addInterval1():
    """
    Test adding Intervals that fall on existng keyframes
    """
    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10----------20 
    # source:             10----------20
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf4 = KF( 20, 35. )
    i1 = Interval(kf2, kf4)
    val = a.addInterval( i1 )
    assert val == True

    kf5 = KF( 10, 8. )
    kf6 = KF( 20, 28. )
    def f(v):
        return v

    i2 = Interval(kf5, kf6, generator=f)
    val = a.addInterval( i2 )
    assert val == True

    assert len(a.intervals) == 2
    assert i1.valGen.generator == i2.valGen.generator
    assert len(a.keyframes) == 3
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf4
    assert a.keyframes[1].getValue() == kf5.getValue()
    assert a.keyframes[2].getValue() == kf6.getValue()


def test_006addInterval1():
    """
    Test adding Intervals that fall on existng keyframes
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10          20-----------30 
    # source:             10----------20
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    kf4 = KF( 30, 35. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == True

    kf5 = KF( 10, 8. )
    kf6 = KF( 20, 28. )
    val = a.addInterval( Interval(kf5, kf6) )
    assert val == True
    
    assert len(a.intervals) == 3
    assert len(a.keyframes) == 4
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[1].getValue() == kf5.getValue()
    assert a.keyframes[2] == kf3
    assert a.keyframes[2].getValue() == kf6.getValue()
    assert a.keyframes[3] == kf4


def test_007addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10
    # source:  0---6
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True

    kf3 = KF( 0, 8. )
    kf4 = KF( 6, 28. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 1
    assert len(a.keyframes) == 2
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    

def test_008addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10
    # source:  0----------------16
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True

    kf3 = KF( 0, 8. )
    kf4 = KF( 16, 28. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 1
    assert len(a.keyframes) == 2
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    

def test_009addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :      10----------20
    # source:  5---------------20
    kf1 = KF( 10, 0. )
    kf2 = KF( 20, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True

    kf3 = KF( 5, 8. )
    kf4 = KF( 20, 28. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 1
    assert len(a.keyframes) == 2
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    

def test_010addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10
    # source:      5------10
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True

    kf3 = KF( 5, 8. )
    kf4 = KF( 10, 28. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 1
    assert len(a.keyframes) == 2
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    

def test_011addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10----------20
    # source:                 12---16
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    val = a.addInterval( Interval(kf2, kf3) )
    assert val == True

    kf4 = KF( 12, 8. )
    kf5 = KF( 16, 28. )
    val = a.addInterval( Interval(kf4, kf5) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 2
    assert len(a.keyframes) == 3
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf3
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    assert a.keyframes[2].getValue() == kf3.getValue()
    

def test_012addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10----------20
    # source:        8--------16
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    val = a.addInterval( Interval(kf2, kf3) )
    assert val == True

    kf4 = KF( 8, 8. )
    kf5 = KF( 16, 28. )
    val = a.addInterval( Interval(kf4, kf5) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 2
    assert len(a.keyframes) == 3
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf3
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    assert a.keyframes[2].getValue() == kf3.getValue()
    

def test_013addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10         20----------30
    # source:        8--------16
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    kf4 = KF( 30, 30. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == True

    kf5 = KF( 8, 8. )
    kf6 = KF( 16, 28. )
    val = a.addInterval( Interval(kf5, kf6) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 2
    assert len(a.keyframes) == 4
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf3
    assert a.keyframes[3] == kf4
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    assert a.keyframes[2].getValue() == kf3.getValue()
    assert a.keyframes[3].getValue() == kf4.getValue()
    

def test_014addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10         20----------30
    # source:                   18--------26
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    kf4 = KF( 30, 30. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == True

    kf5 = KF( 18, 8. )
    kf6 = KF( 26, 26. )
    val = a.addInterval( Interval(kf5, kf6) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 2
    assert len(a.keyframes) == 4
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf3
    assert a.keyframes[3] == kf4
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    assert a.keyframes[2].getValue() == kf3.getValue()
    assert a.keyframes[3].getValue() == kf4.getValue()
    

def test_015addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10         20----------30
    # source:         8------------------26
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    kf4 = KF( 30, 30. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == True

    kf5 = KF( 8, 8. )
    kf6 = KF( 26, 26. )
    val = a.addInterval( Interval(kf5, kf6) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 2
    assert len(a.keyframes) == 4
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf3
    assert a.keyframes[3] == kf4
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    assert a.keyframes[2].getValue() == kf3.getValue()
    assert a.keyframes[3].getValue() == kf4.getValue()
    

def test_016addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0----------10    15     20----------30
    # source:                12---18
    kf1 = KF( 0, 0. )
    kf2 = KF( 10, 5. )
    val = a.addInterval( Interval(kf1, kf2) )
    assert val == True
    kf3 = KF( 20, 30. )
    kf4 = KF( 30, 30. )
    val = a.addInterval( Interval(kf3, kf4) )
    assert val == True

    kf5 = KF( 15, 8. )
    a.addKeyframe( kf5 )

    kf6 = KF( 8, 8. )
    kf7 = KF( 26, 26. )
    val = a.addInterval( Interval(kf6, kf7) )
    assert val == False # fails to add
    
    assert len(a.intervals) == 2
    assert len(a.keyframes) == 5
    assert a.keyframes[0] == kf1
    assert a.keyframes[1] == kf2
    assert a.keyframes[2] == kf5
    assert a.keyframes[3] == kf3
    assert a.keyframes[4] == kf4
    assert a.keyframes[0].getValue() == kf1.getValue()
    assert a.keyframes[1].getValue() == kf2.getValue()
    assert a.keyframes[2].getValue() == kf5.getValue()
    assert a.keyframes[3].getValue() == kf3.getValue()
    assert a.keyframes[4].getValue() == kf4.getValue()


def test_017intervalValGen():
    """
    test that a KF with a different ValueObject overwrites the ValueObject
    when a keyframe is added
    """
    a = Actions()

    kf1 = KF( 0, 0. )

    class Object:
        def __init__(self, value):
            self.value = value
        def get(self):
            return self.value
        def set(self, value):
            self.value = value

    obj = Object(5)


    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))
    assert len(actor.actions.keyframes)==1
    kf0 = actor.actions.keyframes[0]
    assert kf0.getValue() == 5
    assert isinstance(kf0.valueObject, ValueObject)

    autoValue = AutoCurrentValueObject(actor)
    kf1 = KF(0, autoValue)
    actor.addKeyframe( kf1 )
    assert len(actor.actions.keyframes)==1
    assert kf0.getValue() == 5
    assert isinstance(kf0.valueObject, AutoCurrentValueObject)
    obj.set(22)
    assert kf0.getValue() == 22


def test_018intervalValGen():
    """
    test that a KF with a different ValueObject overwrites the ValueObject
    when an Interval is added
    """
    a = Actions()

    kf1 = KF( 0, 0. )

    class Object:
        def __init__(self, value):
            self.value = value
        def get(self):
            return self.value
        def set(self, value):
            self.value = value

    obj = Object(5)

    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))
    assert len(actor.actions.keyframes)==1
    kf0 = actor.actions.keyframes[0]
    assert kf0.getValue() == 5
    assert isinstance(kf0.valueObject, ValueObject)

    kf1 = KFAutoCurrentValue(0, actor)
    kf2 = KF( 10, 20)
    actor.addIntervals( [(kf1, kf2)] )
    assert len(actor.actions.keyframes)==2
    kf = actor.actions.keyframes[0]
    assert isinstance(kf.valueObject, AutoCurrentValueObject)
    assert kf.getValue() == 5
    obj.set(22)
    assert kf.getValue() == 22


def test_020addInterval1():
    """
    Test adding Intervals that fall on existng Intervals
    """

    a = Actions()

    # add intervals from 0-10 and 20-30
    # dest. :  0
    # source:  0---6
    kf1 = KF( 0, 0. )
    val = a.addKeyframe( kf1 )
    assert val == kf1

    class Object:
        def __init__(self, value):
            self.value = value
        def get(self):
            return self.value
        def set(self, value):
            self.value = value

    obj = Object(5)

    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))

    kf2 = KFAutoCurrentValue( 0, actor )
    kf3 = KF( 6, 28. )
    inter = Interval(kf2, kf3)
    val = a.addInterval( inter )
    assert val == True

    assert len(a.intervals) == 1
    assert len(a.keyframes) == 2
    assert a.keyframes[0] == kf2
    assert a.keyframes[1] == kf3
    assert kf2.rightInterval == inter
    assert kf3.leftInterval == inter


def test_021addAction1():
    """
    test adding a new action to an existing one
    """

    act1 = Actions()
    kf1 = KF(0, 10)
    act1.addKeyframe(kf1)
    
    kf2 = KF( 0, (0,0,0) )
    kf3 = KF( 20,(10, 0,0) )
    i1 = Interval(kf2, kf3 )
    act2 = Actions()
    act2.addInterval(i1)
    act1.addActionsAt(act2, 0)

    assert len(act1.keyframes) == 2
    assert act1.keyframes[0].pos == 0
    assert act1.keyframes[1].pos == 20


def checkIntervals(actions, ind1, ind2):
    # check keyframes of actions.intervals[ind1:ind2]
    intervals = actions.intervals
    kfs = actions.keyframes
    kfind = kfs.index(intervals[ind1].kf1)
    for i, inter in enumerate(actions.intervals[ind1:ind2]):
        kf1 = kfs[kfind]
        kf2 = kfs[kfind + 1]
        print "checkIntervals:", kf1.pos, kf2.pos
        assert inter.kf1 == kf1
        assert inter.kf2 == kf2
        assert kf1.rightInterval == inter
        assert kf2.leftInterval == inter
        kfind = kfind + 1
        

def test_022addAction2():
    """
    test adding an action (that will be a copy of existing keyframes and intervals) at a specified position
    """
    # construct first action
    # add intervals from 0-10 and 10-20
    # dest. :  0----------10----------20    
    a1 = Actions()
    kf1 = KF( 0, 25. )
    kf2 = KF( 10, 35. )
    i1 = Interval(kf1, kf2)
    kf3 = KF( 10, 35. )
    kf4 = KF( 20, 45. )
    i2 = Interval(kf3, kf4)
    a1.addInterval(i1)
    a1.addInterval(i2)

    # create a second actions with keyframes and intervals from the first one and add it  at position 30
    # dest :  0----------10----------20
    # source: 0----------10----------20
    # result: 0----------10----------20        30----------40----------50 
    a2 = Actions()
    a2.keyframes = a1.keyframes
    a2.intervals = a1.intervals


    assert a1.addActionsAt(a2, 30) == True
    #print a1.keyframes
    #print  map(lambda x: x.pos, a1.keyframes)
    assert len(a1.keyframes) == 6
    assert len(a1.intervals) == 4
    pos = map(lambda x: x.pos, a1.keyframes)
    assert pos == [0,10,20,30,40,50]
    vals = map(lambda x: x.getValue(), a1.keyframes)
    assert vals == [25., 35., 45., 25., 35., 45.] 
    kfs = a1.keyframes
    assert kfs[0].leftInterval == None
    assert kfs[2].rightInterval == None
    checkIntervals(a1, 0,2)
    assert kfs[3].leftInterval == None
    assert kfs[5].rightInterval == None
    checkIntervals(a1, 2,4)


def test_023addAction3():
    
    a1 = Actions()
    kf1 = KF( 0, 25. )
    kf2 = KF( 10, 35. )
    i1 = Interval(kf1, kf2)
    kf3 = KF( 10, 35. )
    kf4 = KF( 20, 45. )
    i2 = Interval(kf3, kf4)
    a1.addInterval(i1)
    a1.addInterval(i2)

    a2 = Actions()
    kf5 = KF( 0, 40. )
    kf6 = KF( 10, 50. )
    i3 = Interval(kf5, kf6)
    a2.addInterval(i3)
    #a2.keyframes = a1.keyframes[:2]
    #a2.intervals = [a1.intervals[0]]
    
    print "actions 2: ", a2.keyframes
    print "actions 2: ", a2.intervals

    print a1.keyframes
    print a1.intervals
    assert a1.addActionsAt(a2, 0)  == True

    print a1.keyframes
    print a1.intervals
    
    assert len(a1.keyframes) == 3
    assert len(a1.intervals) == 2
    assert a1.getValue(0) == 40.
    assert a1.getValue(10) == 50.
    assert a1.getValue(20) == 45.
    checkIntervals(a1, 0, 2)

    assert a1.addActionsAt(a2, 15) == False
    assert a1.addActionsAt(a2, 5) == False


    

def xtest_100getStringRepr():
    # FIXME 
    a = Actions()

    kf1 = KF( 0, 25. )
    kf2 = KF( 10, 35. )
    i1 = Interval(kf1, kf2)
    val = a.addInterval( i1 )

    kf0 = KF(15, 0.)
    a.addKeyframe( kf0 )

    kf3 = KF( 20, 30. )
    kf4 = KF( 30, 35. )
    val = a.addInterval( Interval(kf3, kf4) )

    labels, posStr, actStr, fullStr = a.getStringRepr()
    assert len(labels) == 4
    assert labels[0] == 0
    assert labels[1] == 10
    assert labels[2] == 20
    assert labels[3] == 30

    assert posStr=='                               |    :    |    :    |    :    |'
    assert actStr=='               ACTOR______NAMEx---------x    x    x---------x'

    assert fullStr=="                                         1         2         3\n                               0         0         0         0\n                               |    :    |    :    |    :    |\n               ACTOR______NAMEx---------x    x    x---------x"
    i1.active = False
    labels, posStr, actStr, fullStr = a.getStringRepr()
    assert len(labels) == 4
    assert labels[0] == 0
    assert labels[1] == 10
    assert labels[2] == 20
    assert labels[3] == 30
    print "posStr:"
    print "z%sz"%posStr
    print 'actStr:'
    print "z%sz"%actStr
    print 'fullStr:'
    print "z%sz"%fullStr
    assert posStr=='               |    :    |    :    |    :    |'
    assert actStr=='ACTOR______NAM x.........x    x    x---------x'
    assert fullStr=='                         1         2         3\n               0         0         0         0\n               |    :    |    :    |    :    |\nACTOR______NAM x.........x    x    x---------x'

    # fullStr is
    ##                          1         2         3
    ##                0         0         0         0
    ##                |    :    |    :    |    :    |
    ## ACTOR______NAM x.........x    x    x---------x


    labels, posStr, actStr, fullStr = a.getStringRepr(4, 22)
    assert len(labels) == 3
    assert labels[0] == 4
    assert labels[1] == 10
    assert labels[2] == 20

    assert posStr=='                :    |    :    |  '
    assert actStr=='ACTOR______NAM ......x    x    x--'
    assert fullStr=='                     1         2  \n               4     0         0  \n                :    |    :    |  \nACTOR______NAM ......x    x    x--'
    # fullStr is
    ##                      1         2  
    ##                4     0         0  
    ##                 :    |    :    |  
    ## ACTOR______NAM ......x    x    x--







##     """
##     Test adding an Interval in the middle of an existing one

##                          i1
##     kf1----------------------------------------------kf2
##                          i2
##                  kf3----------------kf4

## results in:

##           i1             i2               new
##     kf1----------kf3----------------kf4--------------kf2
   
##     """
    
##     a = Actions()

##     # add first interval from 10-40
##     kf1 = KF( 10, 25 )
##     kf2 = KF( 40, 35 )
##     i1 = Interval(kf1, kf2)
##     a.addIntervals( [i1] )
##     assert len(a.intervals) == 1
##     a.intervals[0] = i1
##     assert len(a.keyframes) == 2

##     # add second interval from 20-30
##     kf3 = KF( 20, 25 )
##     kf4 = KF( 30, 35 )
##     i2 = Interval(kf3, kf4)
##     a.addIntervals( [i2] )
##     assert len(a.intervals) == 3
##     assert len(a.keyframes) == 4
##     assert a.keyframes[0] == kf1
##     assert a.keyframes[1] == kf3
##     assert a.keyframes[2] == kf4
##     assert a.keyframes[3] == kf2

##     inter = a.intervals[0]
##     assert inter.kf1 == kf1
##     assert inter.kf2 == kf3

##     inter = a.intervals[1]
##     assert inter.kf1 == kf3
##     assert inter.kf2 == kf4

##     inter = a.intervals[2]
##     assert inter.kf1 == kf4
##     assert inter.kf2 == kf2


## def test_006deleteKF():
##     """
##     test deleteing keyframes
##     """
##     a = Actions()
##     kf1 = KF(0, 10)
##     kf2 = KF(10, 0)
##     i1 = Interval(kf1, kf2)
##     a.addIntervals( [i1] )

##     kf3 = KF(15, 0)
##     a.addKeyframe( kf3 )

##     assert len(a.keyframes) == 3

##     # test that we can delete kf3 because it is isolated
##     a.deleteKeyframe(kf3)
##     assert len(a.keyframes) == 2

##     # test that we can NOT delete kf2 because it belong to an interval
##     a.deleteKeyframe(kf2)
##     assert len(a.keyframes) == 2
    
 
## def test_007addInterval():
##     """
##     Test adding an Interval at the beginning of an existing one

##                          i1
##     kf1----------------------------------------------kf2
##               i2
##     kf3----------------kf4

## results in:

##           new                   i1
##     kf3----------------kf4--------------kf2
   
##     """
    
##     a = Actions()

##     # add first interval from 10-40
##     kf1 = KF( 10, 25 )
##     kf2 = KF( 40, 35 )
##     i1 = Interval(kf1, kf2)
##     a.addIntervals( [i1] )
##     assert len(a.intervals) == 1
##     a.intervals[0] = i1
##     assert len(a.keyframes) == 2

##     # add second interval from 20-30
##     kf3 = KF( 20, 25 )
##     kf4 = KF( 30, 35 )
##     i2 = Interval(kf3, kf4)
##     a.addIntervals( [i2] )
##     assert len(a.intervals) == 3
##     assert len(a.keyframes) == 4
##     assert a.keyframes[0] == kf1
##     assert a.keyframes[1] == kf3
##     assert a.keyframes[2] == kf4
##     assert a.keyframes[3] == kf2

##     inter = a.intervals[0]
##     assert inter.kf1 == kf1
##     assert inter.kf2 == kf3

##     inter = a.intervals[1]
##     assert inter.kf1 == kf3
##     assert inter.kf2 == kf4

##     inter = a.intervals[2]
##     assert inter.kf1 == kf4
##     assert inter.kf2 == kf2

## ## def test_004ActionsAddkeyframe2()
## ##     """
## ##     Test adding a KF that fall on nothing or an existing Interval
## ##     """
## ##      kf1 = KF( 10, 25 )
## ##      kf2 = KF( 20, 35 )

## ##      i1 = Interval(kf1, kf2)

## ##      a = Actions()
## ##      A

## def test_008AddAction():
##     kf0 = KF(0, (1,1,1))
##     act1 = Actions()
##     act1.addKeyframe(kf0)
##     kf1 = KF( 0, (0,0,0) )
##     kf2 = KF( 20,(10, 0,0) )
##     i1 = Interval(kf1, kf2 )
##     act2 = Actions()
##     act2.addIntervals([i1])
##     act1.addActionsAt(act2, 0)

##     assert len(act1.keyframes) == 2
##     assert act1.keyframes[0].pos == 0
##     assert act1.keyframes[1].pos == 20



##
##  Authors:  Michel Sanner, Anna Omelchenko May 2007
##
"""
This module implements a variety of interpolator objects.

An interpolator object has values for the first and last positions (firstVal , lastVal).
This object implements the getValue(fraction) which will return
the value corresponding to the the position (a fraction of the range 0.0 ... 1.0)
"""
import numpy
from math import acos, sqrt, sin, cos, pi
from copy import copy
import weakref
import types
import warnings

class BehaviorList:
    """Class containing a list of behavior objects applied sequentially"""
    nbvar = None
    def __init__(self, firstVal, lastVal, actor=None, active=True):
        self.firstVal = None 
        self.lastVal = None 
        self.behaviors = []
        self.configure(firstVal = firstVal, lastVal = lastVal, active = active)
        self.actor= None
        if actor:
            self.actor = weakref.ref(actor)
        self.valueFormat = None
        self.active = active
        self.easeIn = False
        self.easeOut = False
        self.constant = False


    def clone(self):
        instance = self.__class__(self.firstVal, self.lastVal, active = self.active)
        instance.easeIn = self.easeIn
        instance.easeOut = self.easeOut        
        instance.constant = self.constant
        for b in self.behaviors:
            instance.addBehavior(b.clone())
        return instance

    
    def addBehavior(self, behavior = None, behaviorClass = None):
        if behavior:
            assert isinstance(behavior, Interpolator)
            nbvar = behavior.nbvar
        else:
            assert issubclass(behaviorClass, Interpolator)
            nbvar = behaviorClass.nbvar
            behavior =  behaviorClass(self.firstVal, self.lastVal)
        if behavior:
            if self.nbvar is not None:
                if nbvar is not None:
                    if self.nbvar != nbvar:
                        warnings.warn("Cannot add behavior - expected %d number of variables, got %d"%(self.nbvar, nbvar))
                        return False
            self.behaviors.append(behavior)
            if behavior.valueFormat is not None:
                self.valueFormat = behavior.valueFormat
            behavior.behaviorList = weakref.ref(self)
        return True


    def removeBehavior(self, behavior):
        try:
            self.behaviors.remove(behavior)
            del behavior.behaviorList
        except ValueError:
            pass


    def getValue(self, fraction, interval, **kw):

        #print "behaviorList,easeIn-Out",  self.easeIn , self.easeOut 
        if len(self.behaviors)==0:
            return None
        else:
            value = self.behaviors[0].getValue(fraction, interval)
            for b in self.behaviors[1:]:
                value = b.getValue(fraction, interval, value=value)
            return value


##         if len(self.behaviors)==0:
##             return None
##         else:
##             # alter fraction based on easeIn easeOut
##             if not self.constant and (self.easeIn or self.easeOut):
##                 #print 'EASE 1 ', fraction, 
##                 if self.easeIn and self.easeOut :
##                     #print "easeIn and easeOut"
##                     fraction = (sin(pi*.5*(fraction*2-1))+1)*.5
##                 elif self.easeIn:
##                     #print "easeIn"
##                     fraction = 1.0 - cos(pi*.5 * fraction)
##                 elif self.easeOut:
##                     #print "easeOut"
##                     fraction = sin(pi*.5 * fraction)
##                 #print fraction

##             value = self.behaviors[0].getValue(fraction, interval)
##             for b in self.behaviors[1:]:
##                 value = b.getValue(fraction, interval, value=value)
##             if not (self.constant and (self.easeIn or self.easeOut)):
##                 print 'noeaseinout', fraction, value
##             #print 'in behavior', fraction, value, len(self.behaviors)

##             # for constant interpolation alter value based on easeIn easeOut

##             # constant is used for rotation because each time we get here]
##             # the value get incremented by the same amount (self.firstVal)

##             # the multiplicative factors (2 and 4/3) are comouted so that
##             # the integral remians the same (i.e.) if no ease in or out
##             # we get self.firstVal * nbFrames to reach the desired rotation
##             # if we have ease in we have to accelerate linearly up to the
##             # middle of the interval and then stay constant in such a way
##             # that the sum of all increments is equal to self.firstVal*nbFrames

##             # for some reason we are always short by 0.5*(4/3) on ease in only
##             # and long by 0.5*(4/3) on ease out only. So we special case and
##             # add .5 to ease in on the last frame and start with .5 on ease out
##             if self.constant and (self.easeIn or self.easeOut):
##                 fv = self.firstVal
##                 valtype = type(fv)
##                 if valtype == types.ListType or valtype == types.TupleType:
##                     fv = numpy.array(fv)

##                 if self.easeIn and self.easeOut :
##                     if fraction<0.5:
##                         value = 2*fv*2*fraction
##                     else:
##                         value = 2*fv*2*(1.-fraction)
##                     print "easeInOut", fraction, value, fv
##                 elif self.easeIn:
##                     #if fraction==1.0:
##                     #    value = 1.5*(4./3.)*fv
##                     #else:
##                     if fraction<=0.5:
##                         #value = 2*(4./3.)*fv*fraction
##                         value = dy*fraction
##                     else:
##                         value = h*fraction
##                         #value = (4./3.)*fv
##                     print "easeIn", fraction, value, fv
##                 elif self.easeOut:
##                     #if fraction==0.0:
##                     #    value = .5*(4./3.)*fv*(1.- fraction)
##                     #else:
##                     if fraction>0.5:
##                         value = 2*(4./3.)*fv*(1.- fraction)
##                     else:
##                         value = (4./3.)*fv
##                     print "easeOut", fraction, value, fv
                    
##             return value


    def testKW(self, **kw):
        options = ["firstVal", "lastVal", "active"]
        for k in kw.keys():
            assert k in options
            
    def configure(self, **kw):
        self.testKW(**kw)
        firstVal = kw.get("firstVal")
        lastVal = kw.get("lastVal")
        if firstVal is not None and lastVal is not None:
           self.firstVal = firstVal
           self.lastVal = lastVal
           for b in self.behaviors:
                if b is not None:
                    b.configure(firstVal=firstVal, lastVal=lastVal)
        elif firstVal is not None:
            self.firstVal = firstVal
            for b in self.behaviors:
                if b is not None:
                    b.configure(firstVal=firstVal)

        elif lastVal is not None:
            self.lastVal = lastVal
            for b in self.behaviors:
                if b is not None:
                    b.configure(lastVal=lastVal)
        try:
            nbvar1 = len(self.firstVal)
        except:
            nbvar1 = 1
        try:
            nbvar2 = len(self.lastVal)
        except:
            nbvar2 = 1
        self.nbvar = max(nbvar1, nbvar2)
        active = kw.get("active")
        if active is not None:
            self.active = active
            for b in self.behaviors:
                if b is not None:
                    b.configure(active=active)

        
    def formatValue(self, value):
        if self.valueFormat:
            if type(self.valueFormat) == types.MethodType:
                return self.valueFormat(value)
            else:
                return self.valueFormat%value
        else:
            return str(value)

    def __call__(self, fraction, interval, **kw):
        return self.getValue( *(fraction, interval), **kw)
    

class Interpolator:
    """
Base class for all interpolators.

The getValue method has to be implemented to compute the appropriate value
"""
    nbvar = 1
    varnames = ['variable 1']
    valueFormat = None
    
    def __init__(self, firstVal, lastVal, interpolation = 'linear', active = True):
        self.action = None # will be a weakref to the action this interpolator
                           # is associated with
        self.firstVal = None
        self.lastVal = None
        self.interpolation = interpolation
        self.behaviorList = None
        self.configure(firstVal = firstVal, lastVal = lastVal,
                       interpolation = interpolation, active = active)
        

    def clone(self):
        return self.__class__(self.firstVal, self.lastVal,
                              self.interpolation, active = self.active )

    def testKW(self, **kw):
        options = ["firstVal", "lastVal", "active", "interpolation", "interpolators"]
        for k in kw.keys():
            assert k in options
    
    def configure(self, **kw):

        #print "Interpolator, configure", kw
        self.testKW(**kw)
        firstVal = kw.get("firstVal")
        updateVR = False
        if firstVal is not None:
            self.firstVal = self.cast(firstVal)
            updateVR = True
        lastVal = kw.get("lastVal")
        if lastVal is not None:
            self.lastVal = self.cast(lastVal)
            updateVR = True
        interp = kw.get("interpolation")
        if interp is not None:
            self.interpolation = interp
        if updateVR:
            if self.firstVal is not None and self.lastVal is not None:
                self.valueRange = self.getValueRange()
        try:
            nbvar1 = len(self.firstVal)
        except:
            nbvar1 = 1
        try:
            nbvar2 = len(self.lastVal)
        except:
            nbvar2 = 1
        self.nbvar = max(nbvar1, nbvar2)
        active = kw.get("active")
        if active is not None:
            self.active = active


    def getValueRange(self):
        #print "getValueRange of Interp."
        #return self.lastVal - self.firstVal
        return None


    def getValue(self, fraction, interval, **kw):
        """This method has to return the right value for a fraction ranging
        between 0.0 and 1.0.
        For fraction 0.0 or less it should return self.startValue
        For fraction 1.0 or more it should return self.startValue
        """
        return fraction
        

    def cast(self, value):
        # function used by getValue to return the proper type
        return value


    def ease(self, fraction, interval):
        # computes new fraction for easeIn/easeOut.

        easeIn = False
        easeOut = False
        if self.behaviorList is not None:
            bl = self.behaviorList()
            easeIn = bl.easeIn
            easeOut = bl.easeOut
        if not easeIn and not easeOut:
            return fraction

        # a and b - times to control easeIn and easeOut (in range 0...1)
        # easeIn occurs from time 0 to time a.
        # easeOut occurs from time b to time 1.

        if not interval.data.has_key('easeInEnd'):
            easeInEnd = interval.data['easeInEnd'] = 0.3
        else:
            easeInEnd =  interval.data['easeInEnd']

        if not interval.data.has_key('easeOutStart'):
            easeOutStart = interval.data['easeOutStart'] = 0.7
        else:
            easeOutStart = interval.data['easeOutStart']
        if easeIn and easeOut:            
            a = easeInEnd
            b  = easeOutStart
            #print "easeIn-Out"
        elif easeIn:
            a = easeInEnd 
            b = 1.0
            #print "easeIn"
        else : #easeOut
            #print "easeOut"
            a = 0.0
            b = easeOutStart
            
        v0 = 2/(1+b-a)          # constant velocity attained
        if fraction<=a:
            if a==0:
                return 0
            else:
                d = v0*fraction*fraction/(2*a);
        else:
            d = v0*a/2
            if fraction<=b:
                d += (fraction-a)*v0
            else:
                d += (b-a)*v0
                d += (fraction-fraction*fraction/2-b+b*b/2)*v0/(1-b)

        return d # distance from 0 in range(0, 1)




    def __call__(self, fraction, interval, **kw):
        return self.getValue( *(fraction, interval), **kw)

##     def formatValue(self, value):
##         if self.valueFormat:
##             return self.valueFormat%value
##         else:
##             return str(value)


class noise3F(Interpolator):
    nbvar = 3
    def getValue(self, fraction, interval, value=None, **kw):
        if value is not None:
            if fraction <= 0.0 or fraction >= 1.0:
                return value
            else:
                from random import random as r
                return [value[0]+r(), value[1]+r(), value[2]+r()]


    
class speedUp3F(Interpolator):
    nbvar = 3
    def getValue(self, fraction, interval, value=None, **kw):
        bl = self.behaviorList()
        v1 = bl.firstVal
        v2 = bl.lastVal
        v = [v2[0]-v1[0],v2[1]-v1[1],v2[2]-v1[2]]
        #print fraction, v1, v2, v
        return [v1[0]+fraction*v[0], v1[1]+fraction*v[1], v1[2]+fraction*v[2]]



class noise(Interpolator):
    nbvar = None
    def getValue(self, fraction, interval, value=None, **kw):
        if value is not None:
            if fraction <= 0.0 or fraction >= 1.0:
                return value
            else:
                from random import random as r
                return  map(lambda x: x+r(), value)

            

            

class ScalarInterpolator(Interpolator):
    """ class for performing linear interpolations on scalar values """
    nbvar = None
    
    def __init__(self, firstVal, lastVal, interpolation='linear', active=True):
        
        Interpolator.__init__(self, firstVal, lastVal,
                              interpolation = interpolation, active=active)


    def getValueRange(self):
        #print "getValueRange of Interp."
        if self.lastVal is not None and self.firstVal is not None:
            return self.lastVal - self.firstVal
        else:
            return None

    def getValue(self, fraction, interval, **kw):
        """return the value for a fraction of the range"""
        if self.firstVal is None or self.lastVal is None:
            return None
        if fraction <= 0.0:
            return self.cast(self.firstVal)
        elif fraction >= 1.0:
            return self.cast(self.lastVal)
        else:
            if self.interpolation=='linear':
                fraction = self.ease(fraction, interval)
                return self.cast(self.firstVal +
                                 self.valueRange*fraction)
            else:
                print 'spline interpolation not yet implemented'


class IntScalarInterpolator(ScalarInterpolator):

    def cast(self, value):
        return int(round(value))



class FloatScalarInterpolator(ScalarInterpolator):
    valueFormat = "%.2f"
    
    def cast(self, value):
        return float(value)



class VarScalarInterpolator(Interpolator):
    """
    This class can handle interpolating between a scalar value and a sequence
    of values .
    Example : firstVal = v1, lastVal = [ v21, v22, ..., v2n]

    getValue(fraction) method returns a list containing either 1 or n values for
    specified fraction.
"""
    nbvar = None
    
    def __init__(self, firstVal, lastVal, interpolation='linear', active=True):
        
        Interpolator.__init__(self, firstVal, lastVal,
                              interpolation = interpolation, active=active)

        
    def getValueRange(self):
        # compute value range
        valueRange = []
        try:
            len(self.firstVal)
        except:
            self.firstVal = [self.firstVal]
            if self.behaviorList is not None:
                self.behaviorList().firstVal = [self.firstVal]
        try:
            len(self.lastVal)
        except:
            self.lastVal = [self.lastVal]
            if self.behaviorList is not None:
                self.behaviorList().lastVal = [self.lastVal]
        val1 = self.firstVal
        val2 = self.lastVal
        if len(val1) == 1 and len(val2) > 1:
            val1 = list(val1)*len(val2)
        elif len(val2) == 1 and len(val1) > 1:
            val2 = list(val2)*len(val1)
        vlist = []
        for v1,v2 in zip(val1, val2):
            vlist.append( v2 - v1 )
        return vlist
        
    def getValue(self, fraction, interval, **kw):
        """return the value for a fraction of the range"""
        if self.firstVal is None or self.lastVal is None:
            return None
        if fraction <= 0.0:
            return self.cast(self.firstVal)
        elif fraction >= 1.0:
            return self.cast(self.lastVal)
        else:
            if self.interpolation=='linear':
                fraction = self.ease(fraction, interval)
                if len(self.firstVal) == 1 and len(self.valueRange) > 1:
                    firstVal = list(self.firstVal)*len(self.valueRange)
                else:
                    firstVal = self.firstVal
                return self.cast(map(lambda v1, v2, s=fraction:\
                        v2+s*v1, self.valueRange, firstVal))
            else:
                print 'spline interpolation not yet implemented'



class  IntVarScalarInterpolator(VarScalarInterpolator):
    
    def cast(self, value):
        newval = []
        try:
            len(value)
        except:
            value = [value]
        for v in value:
            newval.append( int(round(v)))
        return newval


        
class  FloatVarScalarInterpolator(VarScalarInterpolator):
    valueFormat = "%.2f"    
    def cast(self, value):
        newval = []
        try:
            len(value)
        except:
            value = [value]
        for v in value:
            newval.append( float(v) )
        return newval


class VectorInterpolator(Interpolator):
    """
    class for performing linear interpolations on vectors of arbitrary length
    Example: firstVal = [v1, v2,....vn], lastVal=[v1, v2,....vn]
    """
    nbvar = None
    
    def __init__(self, firstVal, lastVal, interpolation='linear', active=True):
        
        Interpolator.__init__(self, firstVal, lastVal,
                              interpolation = interpolation, active=active)

        
    def getValueRange(self):
        # compute value range
        #print "getValueRange of vector Interp."
        valueRange = []
        for v1,v2 in zip(self.firstVal, self.lastVal):
            valueRange.append( v2 - v1 )
        return valueRange


        
    def getValue(self, fraction, interval, **kw):
        """return the value for a fraction of the range"""
        if self.firstVal is None or self.lastVal is None:
            return None
        if fraction <= 0.0:
            return self.cast(self.firstVal)
        elif fraction >= 1.0:
            return self.cast(self.lastVal)
        else:
            if self.interpolation=='linear':
                fraction = self.ease(fraction, interval)
                return self.cast(map(lambda x, off, s=fraction:\
                        off+s*x, self.valueRange, self.firstVal))
            else:
                print 'spline interpolation not yet implemented'



class IntVectorInterpolator(VectorInterpolator):
    """Handles interpolation of vectors of integer type"""

    def cast(self, value):
        if isinstance(value, numpy.ndarray):
            newvalue = value.astype('i')
        else:
            newvalue = []
            for val in value:
                newvalue.append(int(round(val)))
        return newvalue



class FloatVectorInterpolator(VectorInterpolator):
    """Handles interpolation of vectors of float type"""
    
    valueFormat = "%.2f"    
    def cast(self, value):
        if isinstance(value, numpy.ndarray):
            newvalue = value.astype('f')
        else:
            newvalue = []
            for val in value:
                newvalue.append(float(val))
        return newvalue

    
class VarVectorInterpolator(Interpolator):
    """
    class for performing linear interpolations on vector sequences of arbitrary length. 
    This can handle interpoalting between an array with shape (1, n) and an array with
    shape (N, n). 
    Example : firstVal = [ [v1, v2,..,vn] ],  lastVal = [ [v11, v12,..,v1n],...,[vN1, vN2,..,vNn] ] 
    """
    nbvar = None
    valueFormat = "%.2f"
    
    def __init__(self, firstVal, lastVal, interpolation='linear', active=True):
        
        Interpolator.__init__(self, firstVal, lastVal, interpolation = interpolation,
                              active=active)

    
    def getValueRange(self):
        # compute value range
        self.firstVal= numpy.array(self.firstVal, 'f')
        self.lastVal= numpy.array(self.lastVal, 'f')
        return self.lastVal - self.firstVal     

            
    def getValue(self, fraction, interval, **kw):
        """return the value for a fraction of the range"""
        if self.firstVal is None or self.lastVal is None:
            return None
        if fraction <= 0.0:
            return self.cast(self.firstVal)
        elif fraction >= 1.0:
            return self.cast(self.lastVal)
        else:
            if len(self.firstVal)==1 and len(self.valueRange)>1:
                # one color at begin and N colors at the end
                firstVal = numpy.array( list(self.firstVal)*len(self.valueRange), "f" )
            else:
                firstVal = self.firstVal
            if self.interpolation=='linear':
                fraction = self.ease(fraction, interval)
                return self.cast(firstVal + self.valueRange*fraction)
            else:
                print 'spline interpolation not yet implemented'

    def cast(self, value):
        if type(value) != numpy.ndarray:
            return numpy.array(value, "f")
        return value.astype("f")


class BooleanInterpolator(Interpolator):
    """class for interpolating values between 0 and 1 to boolean values True or False"""
    
    valueFormat = lambda self, x: str(x == True)
    
    def __init__(self, firstVal, lastVal, interpolation='linear', active=True):
        
        Interpolator.__init__(self, firstVal, lastVal,
                              interpolation = interpolation, active=active)
                            
        
    def getValueRange(self):
        if self.firstVal == True:
            self.firstVal = 1.0
        elif  self.firstVal == False:
            self.firstVal = 0.0
        if self.lastVal == True:
            self.lastVal = 1.0
        elif  self.lastVal == False:
            self.lastVal = 0.0
        return self.lastVal-self.firstVal

   
    def getValue(self, fraction, interval, **kw):
        if self.firstVal is None or self.lastVal is None:
            return None
        if self.firstVal:
            firstVal = 1.0
        else: firstVal = 0.0

        if self.lastVal:
            lastVal = 1.0
        else: lastVal = 0.0
        if fraction <= 0.0:
            if round(firstVal):
                return True
            else:
                return False
        elif fraction >= 1.0:
            if round(lastVal):
                return True
            else:
                return False
          
        else:
            if self.interpolation=='linear':
                if fraction < 1:
                    return {0: False, 1:True}.get(firstVal)
                else:
                    return {0: False, 1:True}.get(lastVal)
            else:
                print 'spline interpolation not yet implemented'




from mglutil.math.rotax import mat_to_quat
from mglutil.math.transformation import Transformation

class RotationInterpolator(Interpolator):
    """
    class for performing quaternion-based interpolations between rotation matrices
    described by 4x4 matrices.
    """
    nbvar = 4
    varnames = ['q1', 'q2', 'q3', 'q4'] # unit quaternion
    valueFormat = "%.2f"    

    def __init__(self, firstVal, lastVal, interpolation='linear', active=True):
        Interpolator.__init__(self, firstVal, lastVal, interpolation = interpolation,
                              active=active)
        self.invertVal = False
        

##     def cast(self, value):
##         t = Transformation(quaternion=value)
##         return t.getRotMatrix(shape=(16,), transpose=1)

        
##     def getValueRange(self):
##         # compute value ranges over all intervals
##         if len(self.firstVal)==16:
##             self.firstVal = mat_to_quat( self.firstVal ) 
##         if len(self.lastVal)==16:
##             self.lastVal = mat_to_quat( self.lastVal ) 
##         v0 = self.firstVal
##         v1 = self.lastVal
##         return (v1[0]-v0[0], v1[1]-v0[1], v1[2]-v0[2], v1[3]-v0[3])
       


##     def getValue(self, fraction, interval):
##         """return the value for a fraction of the range"""
##         if fraction <= 0.0:
##             return self.cast(self.firstVal)
##         elif fraction >= 1.0:
##             return self.cast(self.lastVal)
##         x,y,z,q = self.firstVal
##         rx,ry,rz,rq = self.valueRange
##         return self.cast( (float(x + fraction*rx), float(y + fraction*ry),
##                            float(z + fraction*rz), float(q + fraction*rq)) )

    def configure(self, **kw):
        self.testKW(**kw)
        firstVal = kw.get("firstVal")
        if firstVal is not None:
            assert len(firstVal) == 4
            self.firstVal = firstVal
        lastVal = kw.get("lastVal")
        if lastVal is not None:
            assert len(lastVal) == 4
            self.lastVal = lastVal
        interp = kw.get("interpolation")
        if interp is not None:
            self.interpolation = interp
        if self.behaviorList:
            self.behaviorList().alterFunction = False
        active = kw.get("active")
        if active is not None:
            self.active = active


    def getValue(self, fraction, interval, **kw):
        # quaternion interpolation
        # http://www.euclideanspace.com/maths/algebra/realNormedAlgebra/quaternions/slerp/index.htm
        if self.firstVal is None or self.lastVal is None:
            return None
        qa = self.firstVal
        qb = copy(self.lastVal)
        #qb = self.lastVal
        self.invertVal = False
        if self.behaviorList:
            self.behaviorList().invertVal = False
            
        t = self.ease(fraction, interval)
        cosHalfTheta =  qa[0] * qb[0] + qa[1] * qb[1] + qa[2] * qb[2] + qa[3] * qb[3]
        #print "fraction:", fraction, "cosHalfTheta: ", cosHalfTheta
        #we need cosHalfTheta to be positive - invert the quaternion
        if cosHalfTheta < 0:  # ????
            qb[0] = -qb[0]; qb[1] = -qb[1]; qb[2] = -qb[2]; qb[3] = -qb[3] 
            cosHalfTheata = -cosHalfTheta
            self.invertVal = True
            if self.behaviorList:
                self.behaviorList().invertVal = True

        if abs(cosHalfTheta) >= 1.0:
            qm = [qa[0], qa[1], qa[2], qa[3]]
            return qm
        #Calculate temporary values.
        from math import acos, sqrt, fabs, sin, pi
	halfTheta = acos(cosHalfTheta)
        
	sinHalfTheta = sqrt(1.0 - cosHalfTheta*cosHalfTheta)
	# we could rotate around any axis normal to qa or qb
        if fabs(sinHalfTheta) < 0.001: #fabs is floating point absolute
            qm = [(qa[0] * 0.5 + qb[0] * 0.5), (qa[1] * 0.5 + qb[1] * 0.5),
                  (qa[2] * 0.5 + qb[2] * 0.5), (qa[3]* 0.5 + qb[3] * 0.5)]
            return qm
	ratioA = sin((1 - t) * halfTheta) / sinHalfTheta
	ratioB = sin(t * halfTheta) / sinHalfTheta
	#calculate Quaternion.
	qm = [(qa[0] * ratioA + qb[0] * ratioB), (qa[1] * ratioA + qb[1] * ratioB),
              (qa[2] * ratioA + qb[2] * ratioB), (qa[3] * ratioA + qb[3] * ratioB)]
	return qm


class CompositeInterpolator(Interpolator, object):
    """
    class for aggregating multiple interpolators
    
    the position vector is the same for all interpolators and stored in the
    CompositeInterpolator.
    The values are stored in each interpolator.
    CompositeInterpolator.values is overwriten by a property object collecting
    the values form the interpolators.
    """
    nbvar = None
    varnames = None
    
    def getFirstVal(self):
        val = []  # build list of values
        for inter in self.interpolators:
            val.append(inter.firstVal)
        return val

    
    def getLastVal(self):
        val = []  # build list of values
        for inter in self.interpolators:
            val.append(inter.lastVal)
        return val

    
    def setLastVal(self, val):
        #self.configure(values=values)
        pass


    def setFirstVal(self, val):
        pass

        
    firstVal = property(getFirstVal, setFirstVal)
    lastVal = property(getLastVal, setLastVal)

    
    def __init__(self, firstVal, lastVal, interpolation='linear', interpolators=None,
                 active=True):
        # firstVal - list of values for the first position  of each interpolator
        # lastVal -  lists of values for the last position of each interpolator
        self.interpolators = []
        self.interpolation = interpolation
        self.configure(firstVal = firstVal, lastVal = lastVal, interpolators=interpolators,
                       interpolation = interpolation, active=active)

    def configure(self, **kw):
        self.testKW(**kw)
        interpolators = kw.get("interpolators")
        firstVal = kw.get("firstVal")
        lastVal = kw.get("lastVal")
        if interpolators:
            ninterp = len(interpolators)
            if firstVal is None:
                if not len(self.firstVal):
                    firstVal = []
                    for i in range(ninterp):
                        firstVal.append(None)
                else:
                    firstVal = self.firstVal
            if lastVal is None:
                if not len(self.lastVal):
                    lastVal = []
                    for i in range(ninterp):
                        lastVal.append(None)
                else:
                    lastVal = self.lastVal
            assert len(firstVal)==len(lastVal)==len(interpolators)
            self.interpolators = []
            self.interpConfig = []
            for interpKlass, v1, v2 in zip(interpolators, firstVal, lastVal):
                interpolation = kw.get("interpolation", self.interpolation)
                interp = interpKlass(v1,v2,interpolation = interpolation)
                # replace .configure method with _configure to prevent a user
                # from accessing the interp and modifying a single one
                self.interpolators.append(interp)
                self.interpConfig.append(interp.configure)
                interp.configure = None
        else:
            if lastVal != None and firstVal != None:
                assert len(firstVal)==len(lastVal)==len(self.interpolators)
                for i in range(len(self.interpolators)):
                    self.interpConfig[i](firstVal = firstVal[i], lastVal = lastVal[i])
            elif firstVal is not None:
                assert len(firstVal) == len(self.interpolators)
                for i in range(len(self.interpolators)):
                    self.interpConfig[i](firstVal = firstVal[i])
            elif lastVal is not None:
                assert len(lastVal) == len(self.interpolators)
                for i in range(len(self.interpolators)):
                    self.interpConfig[i](lastVal = lastVal[i])
        interpolation = kw.get("interpolation")
        if interpolation is not None:
             for i in range(len(self.interpolators)):
                    self.interpConfig[i](interpolation = interpolation)
             self.interpolation = interpolation
        active = kw.get("active")
        if active is not None:
            self.active = active
        

    def getValue(self, fraction, interval, **kw):
        result = []
        for interp in self.interpolators:
            result.append(interp.getValue(fraction, interval))
        return result


class FunctionInterpolator(Interpolator):
    """
    class for performing interpolations using functions
    """
    def __init__(self, firstVal=None, lastVal=None, function=None,interpolation='linear',
                 active=True):
        """
        Constructor for an interpolator object using a function.  This object has to
        implement a getValue(fraction) method which returns a value corresponding
        to the given fraction.
        
        arguments:
        function:          function or method to called at each time step.
                           The function can be a function or a 3-tuple
                           (func, *args, **kw). It will be called using
                           func(*(value,)+args, **kw)
        firstVal, lastVal: range of x values to be mapped to the range of frames of
                           the action this interpolator is associated with.

        """
        self.valueRange = None
        Interpolator.__init__(self, firstVal, lastVal, interpolation = interpolation,
                              active=active)
        self.configure(function = function)


    def getValueRange(self):
        #print "getValueRange of Interp."
        return self.lastVal - self.firstVal
        
        
    def configure(self, **kw):
        firstVal = kw.get("firstVal")
        updateVR = False
        if firstVal is not None:
            self.firstVal = firstVal
            updateVR = True
        lastVal = kw.get("lastVal")
        if lastVal is not None:
            self.lastVal = lastVal
            updateVR = True
        function = kw.get("function")
        if function:
            self.funcTuple = self.checkFunction(function)
        interp = kw.get("interpolation")
        if interp is not None:
            self.interpolation = interp
        if updateVR:
            if self.firstVal != None and self.lastVal != None:
                self.valueRange = self.getValueRange()
        active = kw.get("active")
        if active is not None:
            self.active = active

        
    def checkFunction(self, function):
        # check that functionTuple is of form (func, (), {})
        try:
            f, args, kw = function
            assert callable(f)
            assert isinstance(args, tuple)
            assert isinstance(kw, dict)
        except TypeError:
            assert callable(function)
            f, args, kw = function, (), {}
        return f, args, kw


    def getX(self, fraction):
        if self.valueRange:
            return self.valueRange*fraction
        else:
            return None

    
    def getValue(self, fraction, interval, **kw):
        if self.valueRange:
            fraction = self.firstVal+self.valueRange*fraction
        f, args, kw = self.funcTuple
        return self.cast(f( *(fraction,)+args, **kw))



class ReadDataInterpolator(Interpolator):
    
    def __init__(self, firstVal=None, lastVal=None, function=None, nsteps=0,
                 interpolation='linear', active=True):
        self.valueRange = None
        self.function = None
        
        Interpolator.__init__(self,  firstVal, lastVal, interpolation = interpolation,
                              active=active)
        self.nsteps = nsteps
        self.configure(function = function)
        
        

    def getValueRange(self):
        return None

    def clone(self):
        return self.__class__(self.firstVal, self.lastVal, self.function,
                              nsteps = self.nsteps,
                              interpolation=self.interpolation,
                              active = self.active )


    def configure(self, **kw):
        # if firstVal or lastVal get reconfigured - we will add those values to the
        # range of values returned by function(step)
        #print "ReadDataInterpolator, configure", kw
        
        nsteps = kw.get("nsteps")
        if nsteps is not None:
            if nsteps != self.nsteps:
                self.nsteps = nsteps
        function =  kw.get("function")
        if function is not None:
            if type(function).__name__ in ('ndarray', 'list'):
                self.function = function
                self.firstVal = copy(function[0])
                self.lastVal = copy(function[-1])
                self.configure(nsteps = len(function))
            else:
                assert callable(function)
                self.function = function
                self.firstVal = function(0)
                self.lastVal = function(self.nsteps-1)

        firstVal = kw.get("firstVal")
        if firstVal is not None:
            self.firstVal = firstVal
            
        lastVal = kw.get("lastVal")
        if lastVal is not None:
            self.lastVal = lastVal

        try:
            nbvar1 = len(self.firstVal)
        except:
            nbvar1 = 1
        try:
            nbvar2 = len(self.lastVal)
        except:
            nbvar2 = 1
        nbvar = max(nbvar1, nbvar2)
        if self.nbvar != nbvar: self.nbvar = nbvar 
            
        interp = kw.get("interpolation")
        if interp is not None:
            self.interpolation = interp
        active = kw.get("active")
        if active is not None:
            self.active = active

            

    def getValue(self, fraction, interval, **kw):
        if self.function is None: return None
        
        nsteps = self.nsteps
        if fraction == 0:
            if self.firstVal is not None:
                #print "fraction:", fraction, "firstVal"
                return self.firstVal
            else:
                step = 0
        elif fraction == 1:
            if self.lastVal is not None:
                #print "fraction:", fraction, "lastVal"
                return self.lastVal
            else:
                step = nsteps -1
        else:
            step = round((nsteps-1)*fraction)
        #print "fraction:", fraction, "step:", step
        if type(self.function).__name__ in ('ndarray', 'list'):
            return self.function[int(step)]
        else:
            return self.function(int(step))


from math import sqrt, cos, sin, acos
import numpy

def matToQuaternion(mat):
    # converts rotation matrix to quaternion
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
    trace = 1 + mat[0] + mat[5] + mat[10]
    if trace > 0.00000001:
      S = sqrt(trace) * 2
      X = ( mat[9] - mat[6] ) / S
      Y = ( mat[2] - mat[8] ) / S
      Z = ( mat[4] - mat[1] ) / S
      W = 0.25 * S
    else:
        if  mat[0] > mat[5] and mat[0] > mat[10]:      
            S  = sqrt( 1.0 + mat[0] - mat[5] - mat[10] ) * 2
            X = 0.25 * S
            Y = (mat[4] + mat[1] ) / S
            Z = (mat[2] + mat[8] ) / S
            W = (mat[9] - mat[6] ) / S
        elif mat[5] > mat[10] : 
            S  = sqrt( 1.0 + mat[5] - mat[0] - mat[10] ) * 2
            X = (mat[4] + mat[1] ) / S
            Y = 0.25 * S
            Z = (mat[9] + mat[6] ) / S
            W = (mat[2] - mat[8] ) / S
        else:		       
            S  = sqrt( 1.0 + mat[10] - mat[0] - mat[5] ) * 2
            X = (mat[2] + mat[8] ) / S
            Y = (mat[9] + mat[6] ) / S
            Z = 0.25 * S
            W = (mat[4] - mat[1] ) / S
    

      #The quaternion is then defined as:
      #Q = | X Y Z W |
##    #normalize:
##    n = sqrt(X*X + Y*Y + Y*Y + W*W);
##    X /= n
##    Y /= n
##    Z /= n
##    W /= n
    return [X, Y, Z, W]



def quatToMatrix(q):
    # converts quaternion to matrix
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToMatrix/index.htm
    x, y,z, w = q
    sqw = w*w
    sqx = x*x
    sqy = y*y
    sqz = z*z

    #invs (inverse square length) is only required if quaternion is not already normalised
    invs = 1 / (sqx + sqy + sqz + sqw)
    m = numpy.zeros((4,4), 'f')
    m[0][0] = ( sqx - sqy - sqz + sqw)*invs  #since sqw + sqx + sqy + sqz =1/invs*invs
    m[1][1] = (-sqx + sqy - sqz + sqw)*invs 
    m[2][2] = (-sqx - sqy + sqz + sqw)*invs 
    
    tmp1 = x*y
    tmp2 = z*w
    m[1][0] = 2.0 * (tmp1 + tmp2)*invs 
    m[0][1] = 2.0 * (tmp1 - tmp2)*invs 
    
    tmp1 = x*z
    tmp2 = y*w
    m[2][0] = 2.0 * (tmp1 - tmp2)*invs 
    m[0][2] = 2.0 * (tmp1 + tmp2)*invs 
    tmp1 = y*z
    tmp2 = x*w
    m[2][1] = 2.0 * (tmp1 + tmp2)*invs 
    m[1][2] = 2.0 * (tmp1 - tmp2)*invs
    mat = m.flatten()
    mat[-1] = 1.
    return mat




if __name__=='__main__':
    ip = IntScalarInterpolator(0,10)

    print ip.getValue(0.3, None)
    print ip.getValue(0.34, None)
    print ip.getValue(0.35, None)
    print ip.getValue(0.36, None)
    print "---------------"
    
    ip1 = FloatScalarInterpolator(7., 10.)
    print ip1.getValue(0.3, None)
    print ip1.getValue(0.34, None)
    print ip1.getValue(0.35, None)
    print ip1.getValue(0.36, None) 
    print "---------------"
    
    ip2 = FloatVectorInterpolator( [1,1,1], [10,10,10])
    print ip2.getValue(0.5, None)
    print ip2.getValue(0.6, None)
    print ip2.getValue(0.67, None)
    print "---------------"

    ip3 = IntVectorInterpolator( [1,1,1], [10,10,10])
    print ip3.getValue(0.5, None)
    print ip3.getValue(0.6, None)
    print ip3.getValue(0.67, None)
    print "---------------"
    
    ip4 = VarScalarInterpolator (2, [2,4,6])
    print ip4.getValue(0.5, None)
    print ip4.getValue(0.6, None)
    print ip4.getValue(0.67, None)
    print "---------------"
    
    #ip5 = VarVectorInterpolator([[1,1,1], ], [[6,6,6], [8,8,8], [10,10,10]] )
    ip5 = VarVectorInterpolator( [[6,6,6], [8,8,8], [10,10,10]], [[1,1,1], ] )
    print ip5.getValue(0.5, None)
    print ip5.getValue(0.6, None)
    print ip5.getValue(0.67, None)
    
    print "---------------"
    #firstVal = numpy.identity(4).astype('f')
    #from mglutil.math.rotax import rotax
    #import math
    #rot = rotax((0,0,0), (0,1,0), math.pi)
    #lastVal = numpy.dot(firstVal, rot)
    #ip6 = RotationInterpolator(firstVal.ravel(), lastVal.ravel())
    #print ip6.getValue(0, None)
    #print ip6.getValue(0.3, None )
    #print ip6.getValue(0.5, None)
    #print "---------------"
    
    ip7 = CompositeInterpolator(interpolators = (FloatVectorInterpolator, FloatScalarInterpolator),
                                firstVal =([1,1,1], 1,), lastVal = ([10,10,10], 10))
    
                               
    print ip7.getValue(0.5, None)
    print ip7.getValue(0.6, None)
    print ip7.getValue(0.67, None)



import weakref
from time import time

from Scenario2.actions import Actions
from Scenario2.actor import Actor
from Scenario2 import MaxKF

from SimPy.Simulation import Process, initialize, activate, simulate, now, hold

class ActorProcess(Process):
    """
An ActorProcess is a SimPY Process object that gets created each time a
simulation is run.
"""
    def __init__(self, actor):
        """
Constructor of the ActorProcess object,

arguments:
    Actor:             An instance of an Action object
"""
        Process.__init__(self)
        self.actor = actor
        

    def execute(self):
        actor = self.actor
        if actor.playing:
            # FIXME we should wait for the first frame with values change
            yield hold, self, 0 # wait for the first frame
            director = actor._maa()
            while True:
                frame = now()
                if not director.moveForward:
                    # director.maxFrame is computed in director.run(), before activate()
                    frame = director.maxFrame - frame

                value = actor.actions.getValue(frame)
##                 value, inter, valGen = actor.getValueAt(frame)
##                 if valGen == None:
##                     if inter == -1: # this is the case when the first keyframe != 0 and
##                                     # the time cursor is over the interval before it -
##                                     # there is no interpolation
##                         value = None
##                 elif valGen != -1:
##                     if not valGen.active: value = None
                if value!='Nothing There':
                    # pre-step callback
                    try:
                        f, args, kw = actor.preStep_cb
                        f( *args, **kw)
                    except TypeError:
                        pass
                    #if actor.scenarioname == "DejaVuScenario":
                    director.needsRedraw = True
                    if actor.printValuesWhenSetting:
                        print "setting frame %d value of actor %s to " % (frame, actor.name), value
                    actor.setValue(value)

                    # post-step callback
                    try:
                        f, args, kw = actor.postStep_cb
                        f( *args, **kw)
                    except TypeError:
                        pass
                # Fixme we should wait for the next frame with values change
                yield hold, self, 1.


                
class RedrawActorProcess(ActorProcess):
    
    def execute(self):
        actor = self.actor
        # FIXME we should wait for the first frame with values change
        yield hold, self, 0 # wait for the first frame

        while True:
            frame = now()

            try:
                f, args, kw = actor.preStep_cb
                f( *args, **kw)
            except TypeError:
                pass
            maa = None
            if actor._maa:
                maa = actor._maa()
                if hasattr(maa, "_maaGroup") and maa._maaGroup is not None:
                    maa = maa._maaGroup()
            if maa and hasattr( maa, 'sortPoly'):
                if maa.sortPoly == 'Once' and frame==0:
                    #print "maa.name:", maa.name, "sortPoly once"
                    opacityGeoms, parents = maa.findOpacityGeoms()
                    for obj in opacityGeoms:
                        #print "opacity geom:", obj.name
                        order = {'+Zsort':1, '-Zsort':-1}.get(maa.sortOrder)
                        obj.sortPoly(order=order)
                    maa.reorderTransparentChildren(parents)

                elif maa.sortPoly == 'Always':
                    #print "maa.name:", maa.name, "sortPoly always"
                    opacityGeoms, parents = maa.findOpacityGeoms()
                    for obj in opacityGeoms:
                        order = {'+Zsort':1, '-Zsort':-1}.get(maa.sortOrder)
                        obj.sortPoly(order=order)
                    maa.reorderTransparentChildren(parents)
                
            if actor.printValuesWhenSetting:
                print "setting frame %d value of actor %s" % (frame, actor.name)
            actor.setValue()
            # post-step callback
            try:
                f, args, kw = actor.postStep_cb
                f( *args, **kw)
            except TypeError:
                pass
            # Fixme we should wait for the next frame with values change
            yield hold, self, 1.



class MultipleActorsActions:
    """
    Container for Actions objects for multiple actors.
    
    This object maintains a list of actors.

    attributes:
    .name: MAA name
    .actors: list of actors
    .currentFrame
    .endFrame:
    .firstPosition: First frame of the MAA relative to MAA start (i.e. 0)
    .lastPosition: last frame relative to MAA start (i.e. 0)

    Methods:
    bool <- AddActionsAt(actor, actions, position=0)
    Bool <- addMultipleActorsActionsAt(maa, position=None)
    names, posLabel, posStr, actStr, fullStr <-getStringRepr(pos1=0, pos2=None)
    None <- setValuesAt(self, frame, off=0, actor=None)
    """

    def __init__(self, name='MAactions', startFlag="after previous"):
        """
        MultiActorActions constructor

        MultiActorActions <- MultiActorActions(name='MAactions')
        """
        
        self.name = name
        self.actors = [] # list of actors

        self._director = None # becomes a weakref to MAADirector
        
        self.currentFrame = 0
        self.endFrame = 0
        self.firstPosition = 0
        self.lastPosition = 0

        self.gui = None # will be set when a gui is created
        self.recordingActors = []  # list of actors with recording flag on
           # i.e. actor.recording!='stopped'
        # FIXME should have a set and delete method to check type etc
        self.afterAnimation = [] # list of functions to be called after
                                 # play ends
        self.scenarios = {}
        self.autotrackDict = {}
        self.redrawActor = None
        self.moveForward = True
        self.maxFrame = 0
        self.needsRedraw = False
        # These are the attributes used by the sequence animator for computing
        # time position of this maa in the list of all maas.
        self.startFlag = startFlag
        self.startOffset = 0
        self.setOrient = False
        self.sortPoly = 'Once'
        self.sortOrder = "-Zsort"
        
        

    def AddActions(self, actor, actions, position=0):
        """
        adding actions for an actor at a given position

        bool <- AddActionsAt(actor, actions, position=0)

        arguments:
          actor: has to be an instance of an Actor Object
          actions: is an instance of an Actions object
          position: start frame for this action relative to begin of MAA
        
        return values
           
        if actor already exists in this MAA the actions are added to this actor
        else a new actor is created in this MAA tand the actions are added to
        this new actor
        
        """

        # FIXME
        # we should return the actor so we know when a new one is created
        # - actor is the actor in the MAA which can is a new actor if the
        #       incomming actor was not found in the MAA
        # - success is True if the actions were added successfully to the actor
        
        assert isinstance(actor, Actor)

        assert isinstance(actions, Actions)
        
        objAnim = self.findActor(actor)
        if objAnim is None:
            actor = self.addActor(actor)
        else:
            actor = objAnim

        success = actor.actions.addActionsAt(actions, position)
        if success:
            self.computeFirstAndLastPosition()

        return actor, success


    def findActor(self, actor):
        """
        find an actor if is exists in the director

        actor <- findActor(actor1)

        actor1 is an instance of an actor
        if there is an actor in the director withthe same name we return this actor
        else we return None
        """
        import types
        if type(actor) == types.StringType:
            name = actor
        else:
            name = actor.name
        for objAnim in self.actors:
            if objAnim.name == name:
                return objAnim
        return None


    def addActor(self, actor, redraw = True):
        """
        add an actor to the director.

        actor <- addActor(actor, redraw=True)
        
        if an actor with that name already exists this actor is return
        else a clone of the actor is added and returned
        """
        objAnim = self.findActor(actor)
        if objAnim is not None:
            return objAnim

        assert isinstance(actor, Actor)
        newActor = actor.clone()
        newActor._maa = weakref.ref(self)
        self.actors.append( newActor )
        lastFrame = newActor.getLastKeyFrame()
        if lastFrame is not None:
            lastFrame = lastFrame.pos
        else:
            lastFrame = 0
        if lastFrame >= self.endFrame:
            self.updateEndFrame(lastFrame)

        newActor.onAddToDirector()
        return newActor


    def addMultipleActorsActionsAt(self, maa, position=None):
        """
        add MultipleActorsActions object to a director at a given position

        Bool <- addMultipleActorsActionsAt(maa, position=None)

        if position is None the maa is added at its current position,
        else position can be an absolute frame at which the first keyframe
        of the maa will be placed.
        The boolean retrun True of the maa was successfully added
        """

        mini, maxi = maa.boundingFrames()

        # check if this MAA can be added
        # it can be added if no existing actor has anything between mini and maxi
        if position is None:
            position = mini
            
        for actor in maa.actors:
            objAnim = self.findActor(actor)
            if objAnim is None:
                continue # there will be no problem adding this actions
            if objAnim.actions.keyframeBetween(position, position + maxi - mini):
                return False

        for actor in maa.actors:
            actions = actor.actions
            actor = self.addActor(actor)
            poskf0 = actions.boundingFrames()[0]
            delta = poskf0 - mini
            val = actor.addActionsAt(actions, position+delta )
            if val is False:
                raise RuntimeError ("failed to add MAA")

        self.computeFirstAndLastPosition()
        return True


    def getStringRepr(self, pos1=0, pos2=None):
        """Return a string description of all actions.

        names, posLabel, posStr, actStr, fullStr <-getStringRepr(pos1=0, pos2=None)
        names is the list of actor names
        See Action.getStringRepr for the description of hte other returned values
        """

        allActionsStr = """"""
        names = []
        nokfnames = []
        nokfStr = """"""
        for actor in self.actors:
            actions = actor.actions

            posLab, posStr, actStr, fullStr = actions.getStringRepr(
                pos1, pos2, actor.name)
            if fullStr.find("No Keyframes")> 0:
                nokfnames.append(actor.name)
                nokfStr += fullStr+'\n'
            else:
                names.append(actor.name)
                if len(allActionsStr)==0:
                    allActionsStr += fullStr+'\n'
                else:
                    allActionsStr += actStr+'\n'
        if len(nokfStr):
            allActionsStr += nokfStr
            names.extend(nokfnames)
        return names, posLab, posStr, actStr, allActionsStr

  
    def afterAnimation_cb(self):
        # subclass this
        pass

    
    def computeFirstAndLastPosition(self):
        mini, maxi = self.boundingFrames()
        self.firstPosition = mini
        self.lastPosition = maxi


    def boundingFrames(self, position=0):
        """
        return first and last KF positions for the object
        """
        mini = MaxKF
        maxi = 0

        for actor in self.actors:
            action = actor.actions
            minia, maxia = action.boundingFrames(position)
            if maxia == MaxKF: continue  # for example , in  'redraw' actor case 
            if minia < mini:
                mini = minia
            if maxia > maxi and maxia < MaxKF:
                maxi = maxia
        #if maxi==0:
        #    maxi = MaxKF
        return mini, maxi


    def addScenario(self, name, scenario):
        self.scenarios[name] = scenario


    def start(self, setDirector = True, title = None):
        if self.gui is None:
            gui = DirectorGUI(self, title = title)
            gui.root.protocol('WM_DELETE_WINDOW', gui.root.withdraw)
            self.gui = gui
            if setDirector:
                for scenario in self.scenarios.values():
                    scenario.setDirector()
        else:
            if title:
                oldtitle = self.gui.root.title()
                if oldtitle != title:
                   self.gui.root.title(title) 
            self.gui.root.deiconify()

    def __repr__(self):
        return self.name


##     def deleteActor(self, actor):
##         # delete this actor and all its actions
##         name = actor.name
##         if self.gui:
##             self.gui.deleteActors()
##         if actor in self.actors:
##             self.actors.remove(actor)
##         if hasattr(actor, "guiVar"):
##             if actor.guiVar:
##                 actor.guiVar.set(0)
##         #if hasattr(actor.object, "animatedProperties"):
##         #    del actor.object.animatedProperties[name]
##         if hasattr(actor.scenario, "animatedProperties"):
##             del actor.scenario.animatedProperties[name]
##         if self.gui:
##             if self.gui.selectedFrames.has_key(name):
##                 self.gui.selectedFrames.pop(name)
##             if self.gui.autotracking:
##                 if self.autotrackDict.has_key(name):
##                     self.autotrackDict.pop(name)
##                     # this will update the dictioanary with a new "clean"
##                     # version of the deleted actor.
##                     self.autotrackDict.update(actor.scenario.getNewActors())
                
##             self.gui.drawActors()
##         ####################FIX THIS ###########
##         if len(actor.kfSetId):
##             for kf, setId in actor.kfSetId.items():
##                 kfset = self.gui.kfSets.get(setId)
##                 if kfset:
##                     kfset.removeKeyFrame(name, kf)
##                     if kfset.getNumberOfFrames() == 1:
##                         # remove the set
##                         self.gui.removeKFSet(kfset)

        

    def getLastFrame(self):
        # returns number of frames in the simulation
        return self.endFrame

    
    def getLastFrameWithChange(self, actors = None):
        # returns the last frame in which a value changes
        if not actors:
            actors = self.actors
        last = None
        for actor in actors:
            #if actor.isFullRange(): continue
            if actor.visible:
                lastFrame = actor.getLastKeyFrame()
                if last is None:
                    last = lastFrame
                else:
                    if lastFrame.pos > last.pos:
                        last = lastFrame
        return last


    def updateEndFrame(self, end, updateGUI=1):
        if not end:
            return
        
        if end > self.endFrame:
            self.endFrame = end
            if self.gui and updateGUI:
                self.gui.setDuration(end)

        else:
            last = self.getLastFrameWithChange()
            if last is not None:
                last = last.pos
            else:
                last = 0
            if end < last+10 : # when recording we want to make sure there are frame to the right
                self.endFrame = last +10
            else:
                self.endFrame = end
            if self.gui and updateGUI:
                self.gui.setDuration(self.endFrame)


    def run(self, forward=True):
        """
        used in DejaVu/scenarioInterface/animationPanels.py AnimationPanel.runMAA
        """
        if len(self.actors)==0:
            #print "no actors"
            return
        initialize()
        redraw = self.redrawActor
        start = 0
        #end = self.endFrame
        self.maxFrame = 0
        maxFrame = self.getLastFrameWithChange()
        if maxFrame is not None:
            self.maxFrame = maxFrame.pos

        end = self.maxFrame
        gui = self.gui
        if gui:
            if not (gui.startFrame == 0 and gui.stopFrame == 1):
                start = gui.startFrame
                end = gui.stopFrame
                self.maxFrame = end + start
        self.moveForward = forward
        if redraw:
            redraw.object.stopAutoRedraw()
            proc = RedrawActorProcess(redraw)
            activate(proc, proc.execute(), at=start, prior=True)
            #print 'activated redraw', 
        for actor in self.actors:
            if actor.name != "redraw":
                proc = ActorProcess(actor)
                activate(proc, proc.execute(), at=start, prior=True)
                #print 'activated ', actor.name, action
        #print 'simulate', self.endFrame
        simulate(until=end)
        #self.currentFrame = self.endFrame
        if forward:
            #self.currentFrame = self.endFrame
            self.currentFrame = end
        else:
           self.currentFrame = start
        # call callback after animation completes
        for f in self.afterAnimation:
            f()
        #self.afterAnimation_cb()
        if self.redrawActor:
            vi = self.redrawActor.object
            vi.startAutoRedraw()


    def _timedDraw(self, director, at):
        director.setValuesAt(at)
        director.viewerRedraw()
        dtime = time() - self._timeStart
        fraction = min(1.0, dtime/self._duration)
        frameT = int(fraction*self._nbFrames)
        self.frameCounter += 1
        at = max(frameT, self.frameCounter)
        #print 'AT', at, director.endFrame
        #print dtime, frameT, self.frameCounter, director.endFrame
        if at < self._nbFrames:
            self._afterFunc(1, self._timedDraw(director, at))
            
        
    def runIn(self, duration, afterFunc):
        """
        used in DejaVu/scenarioInterface/animationPanels.py AnimationPanel.runMAA
        """
        from Scenario2.director import MAADirector
        director = MAADirector()
        director.addMAAat(self, 0)

        self.needsRedraw = False

        director.stopViewersAutoRedraw()

        self._nbFrames = director.endFrame-1
        self._duration = float(duration)
        self._afterFunc = afterFunc
        self.frameCounter = 0
        
        self._timeStart = time()
        self._timedDraw(director, 0)

        del self._nbFrames
        del self._duration
        del self._timeStart
        del self._afterFunc
        
        director.startViewersAutoRedraw()

        
    def findOpacityGeoms(self):
        """a helper function used in setValuesAt(). Finds all opacity actors
        and returns a list of their objects(geometries) and  a dictionary of
        the parents of the 'opacity' geometries"""
        
        # find all opacity actors and the parents of the geometry
        opacityGeoms = []
        parents  = {}
        for a in self.actors:
            if a.name[-7:]=='opacity':
                obj = a.object
                opacityGeoms.append(obj)
                if parents.has_key(obj.parent):
                    parents[obj.parent].append(obj)
                else:
                    parents[obj.parent] = [obj]
        return opacityGeoms, parents


    def reorderTransparentChildren(self, parents):
        """a helper function used in setValuesAt().
        Reorder children of same parent from back to front"""
        
        import numpy
        for parent, objs in parents.items():
            allobjs = []
            for o in parent.children:
                if o.visible:
                    allobjs.append(o)
            zmin = []
            for obj in allobjs:
                mat = obj.GetMatrix()
                p1, p2 = obj.ComputeBB()
                p = [ (p1[0], p1[1], p1[2], 1.),
                      (p2[0], p2[1], p2[2], 1.)] 
                tp = numpy.dot(p, mat.transpose())
                # save closest Z value for this geom
                zmin.append( max(tp[0][2], tp[1][2]) )
                # remove child from parent's list
                parent.children.remove(obj)

            # sort by increasing z
            sortedl = numpy.argsort(zmin)

            # append the children to parent's children list
            # ordered from furthest z to closest
            for i in sortedl:#
                parent.children.append(allobjs[i])


    def setValuesAt(self, frame, off=0, actor=None):
        #print "setValuesAt:", frame, self.name
        firstFrame = False
        if frame == off + self.firstPosition: # first Frame
            firstFrame = True
               
        if actor:
            actors = [actor]
        else:
            actors = self.actors
        self.needsRedraw = False
        for actor in actors:
            actions = actor.actions
            if actor.playing:
                actor.setValueAt(frame, off)
        if hasattr( self, 'sortPoly'):
            if self.sortPoly == 'Once' and firstFrame:
                #print "maa.name:", self.name, "sortPoly once"
                opacityGeoms, parents = self.findOpacityGeoms()
                for obj in opacityGeoms:
                    #print "'Once, opacity geom:", obj.name
                    order = {'+Zsort':1, '-Zsort':-1}.get(self.sortOrder)
                    obj.sortPoly(order=order)
                self.reorderTransparentChildren(parents)

            elif self.sortPoly == 'Always':
                #print "maa.name:", self.name, "sortPoly always"
                opacityGeoms, parents = self.findOpacityGeoms()
                for obj in opacityGeoms:
                    #print "Always, opacity geom:", obj.name
                    order = {'+Zsort':1, '-Zsort':-1}.get(self.sortOrder)
                    obj.sortPoly(order=order)
                self.reorderTransparentChildren(parents)

    def getAllGeoms(self):
        """Returns a list of all geometries driven by maa's actors"""
        geoms = []
        from DejaVu.Geom import Geom
        from DejaVu.scenarioInterface.actor import DejaVuTransformationActor
        
        for actor in self.actors:
            if isinstance(actor, DejaVuTransformationActor):
                continue
            obj = actor.object
            if isinstance(obj, Geom) and obj not in geoms:
                geoms.append(obj)
        return geoms


    def gotoFrame(self, frame, set=True):
        if frame<0:
            frame = 0
        elif frame>self.endFrame:
            frame = self.endFrame
        if frame==self.currentFrame:
            return
        self.currentFrame = frame
        if set:
            self.setValuesAt(frame)
        if self.gui:
            self.gui.placeTimeCursor(frame)

    def configure(self, **kw):
        """ should be implemented by a class the  inherits from MultipleActorsActions"""
        pass
            

##     def createActorFromData(self, obj, propname, scenarioInterface,filename, fields,
##                             data, start = 0, end=-1):
##         # Create a special actor whose valueGenerator returns values from a data 
##         # sequence such as list or numpy array, This data is read from a file (.adat)->
##         # see actor.py/adatFileParser

##         # obj - animatable object
##         # propname - the object's animated property name
##         # scenarioInterface - scenario interface that is n self.scenarios and "knows" how
##         # to create a regular actor for this object (i.e has createActor() method).
##         # start - first keyframe of the actor
##         # end - last keyframe of the actor ( if end == -1, then the number of frames between
##         # first and last keyframes is len(data)-1, or one data step per each keyframe).

##         if data is None:
##             print "createActorFromData: data is None"
##             return

##         assert scenarioInterface in self.scenarios.values()
##         actor = None
##         if data is not None:
##             assert type(data).__name__ in ('ndarray', 'list')
##             nsteps = len(data)
##             firstVal = data[0]
##             lastVal =  data[nsteps-1]
##             #print filename, len(data), data[0], propname
##             # create an actor with ReadDataInterpolator valuegenerator
##             #try:

##             # this will create file actor with two keyframes : "0" and "nsteps-1" 
##             actor = scenarioInterface.createActor(obj, propname, check=False,
##                                                   addToDirector=False,
##                                                   actorData = (filename, data))
##             actor.datafields = fields
##             #except:
##             #    warnings.warn("Failed to create file actor %s" % propname)
##             #    return None
##             if actor:
##                 self.addActor(actor)
##                 #print "start=", start, "end=", end, "nsteps=", nsteps
##                 if end == -1:
##                     end = start + nsteps - 1
##                 if end != nsteps-1:
##                     if self.gui:
##                         #this will move "nsteps-1" keyframe  to "end" keyframe 
##                         self.gui.selectKeyFrame_cb(actor, nsteps-1, newframe=end)
##                 if start != 0:
##                     if self.gui:
##                         # this will move "0" keyframe to "start" keyframe
##                         self.gui.selectKeyFrame_cb(actor,0,newframe=start) 
                    
##         return actor



class MAAGroup:


    def __init__(self, name='MAAGroup', maalist=[], startFlag="after previous"):
        """
        MAAGroup constructor

        MAAGroup <- MAAGroup(name='MAAGroup', maalist=[], startFlag='after previous')
        """
        self.maas = []
        self.actors = []
        self.firstPosition = 0
        self.lastPosition = 0
        self.currentFrame = 0
        self.redrawActor = None
        self.afterAnimation = []
        for maa in maalist:
            self.AddMaa(maa)
        self.name = name
        self._director = None # becomes a weakref to MAADirector
        
        self.gui = None # will be set when a gui is created
        self.scenarios = {}

        self.maxFrame = 0

        # These are the attributes used by the sequence animator for computing
        # time position of this maa in the list of all maas.
        self.startFlag = startFlag
        self.startOffset = 0
        self.editorClass = None
        self.setOrient = False
        self.needsRedraw = False
        self.sortPoly = 'Once'
        self.sortOrder = "-Zsort"

        

    def AddMaa(self, maa):
        mini, maxi = maa.boundingFrames()
        if mini < self.firstPosition:
            self.firstPosition=mini
        if maxi > self.lastPosition:
            self.lastPosition = maxi
        self.maas.append(maa)
        maa._maaGroup = weakref.ref(self)
        for actor in maa.actors:
            if actor.name.startswith('redraw'):
                if not self.redrawActor:
                    self.actors.insert(0, actor)
                    self.redrawActor = actor
            else:
                self.actors.append(actor)
            self.afterAnimation.extend(maa.afterAnimation) 
        

    def findActor(self, actor):
        """
        find an actor if is exists in the director

        actor <- findActor(actor1)

        actor1 is an instance of an actor
        if there is an actor in the director withthe same name we return this actor
        else we return None
        """
        for maa in self.maas:
            _actor = maa.findActor(actor)
            if _actor is not None:
                return _actor
        return None

    def makeActorList(self, maas = None, reverse = False):
        self.actors = []
        self.redrawActor = None
        if maas == None:
            maas = self.maas
        for maa in maas:
            actors = maa.actors
            if reverse: 
                # In the maa.run() method the actors are played (by simulate()) in the reversed order
                # from the one they had beed activated. So, if we want to play actors in
                # the original order , we need to reverse the list of actors before activating each actor.
                actors = maa.actors[:]
                actors.reverse()
            for actor in actors:
                if actor.name.startswith('redraw'):
                    if not self.redrawActor:
                        self.actors.insert(0, actor)
                        self.redrawActor = actor
                else:
                    self.actors.append(actor)


    def getAllGeoms(self):
        geoms = []
        for maa in self.maas:
            _geoms = maa.getAllGeoms()
            for obj in _geoms:
                if obj not in geoms:
                    geoms.append(obj)
        return geoms
                    

    def getStringRepr(self, pos1=0, pos2=None):
        """Return a string description of all actions.

        names, posLabel, posStr, actStr, fullStr <-getStringRepr(pos1=0, pos2=None)
        names is the list of actor names
        See Action.getStringRepr for the description of hte other returned values
        """

        allActionsStr = """"""
        names = []
        for maa in self.maas:
            _names, _posLab, _posStr, _actStr, _allActionsStr=maa.getStringRepr(pos1, pos2)
            names.extend(_names)
            allActionsStr += _allActionsStr 
        return names, allActionsStr

  
    def afterAnimation_cb(self):
        # subclass this
        pass

    
    def computeFirstAndLastPosition(self):
        self.firstPosition = 0
        self.lastPosition = 0
        for maa in self.maas:
            mini, maxi = maa.boundingFrames()
            if mini < self.firstPosition:
                self.firstPosition=mini
            if maxi > self.lastPosition:
                self.lastPosition = maxi


    def boundingFrames(self, position=0):
        """
        return first and last KF positions for the object
        """
        self.computeFirstAndLastPosition()
        return self.firstPosition+position, self.lastPosition+position


    def addScenario(self, name, scenario):
        self.scenarios[name] = scenario


    def run(self, forward=True):
        """
        used in DejaVu/scenarioInterface/animationGUI.py animClip.makeAndRunMAA
        """
        if len(self.maas)==0:
            #print "no actors"
            return
        initialize()
        redraw = self.redrawActor
        actors = self.actors
        afterAnimation = self.afterAnimation
        start = self.firstPosition
        end = self.lastPosition
        if redraw:
            redraw.object.stopAutoRedraw()
            proc = RedrawActorProcess(redraw)
            activate(proc, proc.execute(), at=start, prior=True)
            #print 'activated redraw', 
        for actor in actors:
            if actor == redraw:continue
            proc = ActorProcess(actor)
            activate(proc, proc.execute(), at=start, prior=True)
        simulate(until=end)
        #self.currentFrame = self.endFrame
        if forward:
            #self.currentFrame = self.endFrame
            self.currentFrame = end
        else:
           self.currentFrame = start
        # call callback after animation completes
        for f in afterAnimation:
            f()
        #self.afterAnimation_cb()
        if redraw:
            redraw.object.startAutoRedraw()


    def setValuesAt(self, frame, off=0, actor=None):
        for maa in self.maas:
            maa.setValuesAt(frame, off, actor)

    def getSourceCode(self, varname, indent = 0):
        tabs = " "*indent
        lines = tabs + """from Scenario2.multipleActorsActions import MAAGroup\n"""
        lines += tabs +"""%s = MAAGroup(name='%s', startFlag='%s')\n""" %(varname, self.name, self.startFlag)
        for i, maa in enumerate(self.maas):
            lines +=maa.getSourceCode("_maa%d"%i, indent=indent)
            lines += tabs + """%s.AddMaa(_maa%d)\n""" % (varname, i)
        return lines
        
            
    def configure(self, **kw):
        pass



"""
"""

MaxKF = 999999999


from clipboard import Clipboard

# create a default clipboard object that can be shared accros applications
# using scenario

class ScenarioMAATargets:
    """
    Class holding animator object that can  be targets for sending MAAs to
    """

    def __init__(self):
        self.targetNames = []
        self.targetsByName = {}

    def addAnimator(self, name, target, addingMethod):
        assert callable(addingMethod)
        self.targetNames.append(name)
        self.targetsByName[name] = (target, addingMethod)

    def addToTarget(self, name, maa):
        target, addingMethod = self.targetsByName[name]
        addingMethod(target, maa)

_MAATargets = ScenarioMAATargets()

from mglutil.util.callback import CallBackFunction

def addTargetsToMenu(menu, maa):
    for name in _MAATargets.targetNames:
        cb = CallBackFunction( _MAATargets.addToTarget, name, maa)
        menu.add_command(label='Add to '+name, command=cb)
 
# create a default clipboard object that can be shared accros applications
# using scenario

_clipboard = Clipboard()

#
# create an event handle for scenario events
from mglutil.events import EventHandler

_eventHandler = EventHandler()




##
##  Author Michel F. Sanner May 2007
##

import numpy
class DataType:

    def valide(self, value):
        return True

    def cast(self, data):
        return False, None

    def isEqual(self, val1, val2):
        if val1 == val2:
            return True
        else:
            return False



class FloatType(DataType):

    def valide(self, value):
        return isinstance(value, float)

    def cast(self, data):
        try:
            return True, float(value)
        except:
            return False, None

    def isEqual(self, val1, val2):
        if abs(val1-val2) < 1.e-5:
            return True
        else: return False
        



class IntType(DataType):

    def valide(self, value):
        return isinstance(value, int)

    def cast(self, data):
        try:
            return True, int(value)
        except:
            return False, None



class BoolType(DataType):

    def valide(self, value):
        if isinstance(value, bool):
            return True
        else:
            if value == 1 or value == 0:
                return True
            else:
                return False

    def cast(self, data):
        try:
            return True, bool(value)
        except:
            return False, None


import numpy

class IntVectorType(DataType):
    
    def valide(self, value):
        try:
            len(value)
        except:
            return False
        for v in value:
            if not isinstance(v, int):
                return False
        return True

    
    def cast(self, data):
        try:
            if isinstance(data, numpy.ndarray):
                newdata = data.astype('i')
            else:
                newdata = []
                for val in data:
                    newdata.append(int(val))
            return True, newdata
        except:
            return False, None

    def isEqual(self, val1, val2):
        val1 = numpy.array(val1, 'i')
        val2 = numpy.array(val2, 'i')
        return numpy.alltrue(val1, val2)

        
class FloatVectorType(DataType):
    
    def valide(self, value):
        try:
            len(value)
        except:
            return False
        for v in value:
            if not isinstance(v, float):
                return False
        return True


    def cast(self, data):
        try:
            if isinstance(data, numpy.ndarray):
                newdata = data.astype('f')
            else:
                newdata = []
                for val in data:
                    newdata.append(float(val))
            return True, newdata
        except:
            return False, None

    def isEqual(self, val1, val2):
        val1 = numpy.array(val1, 'f')
        val2 = numpy.array(val2, 'f') 
        d = abs(val1-val2)
        for i, v in enumerate(d):
           if  v > 1.e-5:
               #print "value arrays differ: ", i, val1[i], val2[i]
               return False
        return True



class IntVarType(DataType):
    
    def valide(self, value):
        # valid value can be either a single integer or a sequence of integers 
        try:
            len(value)
        except:
            value = [value]
        if isinstance(value, numpy.ndarray):
            dt = value.dtype.char  
            if dt not in ('i', numpy.int):
                return False
        else:
            for v in value:
                if not isinstance(v, int):
                    return False
        return True

    def cast(self, data):
        try:
            len(data)
        except:
            data = [data]
        try:
            if isinstance(data, numpy.ndarray):
                newdata = data.astype('i')
            else:
                newdata = []
                for val in data:
                    newdata.append(int(val))
            return True, newdata
        except:
            return False, None

    def isEqual(self, val1, val2):
        try:
            len(val1)
        except:
            val1 = [val1]
        try:
            len(val2)
        except:
            val2 = [val2]
        val1 = numpy.array(val1, 'i')
        val2 = numpy.array(val2, 'i')
        if val1.shape != val2.shape:
            return False
        return numpy.alltrue(val1, val2)


class FloatVarType(DataType):
    
    def valide(self, value):
        # valid value can be either a single float or a sequence of floats 
        try:
            len(value)
        except:
            value = [value]
        if isinstance(value, numpy.ndarray):
            dt = value.dtype.char  
            if dt not in ('f', numpy.float):
                return False
        else:
            for v in value:
                if not isinstance(v, float):
                    return False
        return True

    def cast(self, data):
        try:
            len(data)
        except:
            data = [data]
        try:
            if isinstance(data, numpy.ndarray):
                newdata = data.astype('f')
            else:
                newdata = []
                for val in data:
                    newdata.append(float(val))
            return True, newdata
        except:
            return False, None


    def isEqual(self, val1, val2):
        try:
            len(val1)
        except:
            val1 = [val1]
        try:
            len(val2)
        except:
            val2 = [val2]
        val1 = numpy.array(val1, 'f')
        val2 = numpy.array(val2, 'f')
        if val1.shape != val2.shape:
            return False
        d = abs(val1-val2)
        for i, v in enumerate(d):
           if  v > 1.e-5:
               #print "value arrays differ: ", i, val1[i], val2[i]
               return False
        return True



class VarVectorType(DataType):

    def valide(self, value):
        #valid value is a 2D sequence
        if isinstance(value, numpy.ndarray):
            if len(value.shape) != 2:
                return False
        else:
            try:
                len(value)
            except:
                return False
            
            try:
                nelem = len(value[0])
            except:
                return False
            if len(value) > 1:
                for v in value[1:]:
                    try:
                        n = len(v)
                    except:
                        return False
                    if n != nelem:
                        return False
        return True

    def isEqual(self, val1, val2):
        val1 = numpy.array(val1, 'f')
        val2 = numpy.array(val2, 'f')
        if val1.shape != val2.shape:
            return False
        d = abs(val1-val2)
        shape = d.shape
        for i in range(shape[0]):
            for j in range(shape[1]):
                if  d[i][j] > 1.e-5:
                    #print "value arrays differ: val1[%d][%d] = %f, val2[%d][%d] = %f"% (i,j, val1[i][j], i, j, val2[i][j])
                    return False
        return True
            

if __name__=='__main__':
    t = IntVectorType()
    v1 = 3
    assert t.valide(v1)==False
    v2 = [1., 2., 3.]
    assert t.valide(v2) == False
    v3 = [1,2,3]
    assert t.valide(v3) == True
    assert t.cast(v2)[1] == v3
    assert t.cast(v1)[0] == False

    t = FloatVectorType()
    v4 = 3.4
    assert t.valide(v4)==False
    assert t.valide(v3) == False
    assert t.valide(v2) == True
    assert t.cast(v3)[1] == v2
    assert t.cast(v4)[0] == False

    t = IntVarType()
    assert t.valide(v4)==False
    assert t.valide(v1)==True
    assert t.valide(v3) == True
    assert t.valide(v2) == False
    v5 = numpy.array([1,2,3], 'f')
    assert t.valide(v5) == False
    cv = t.cast(v5)
    assert t.valide(cv[1]) == True
    assert t.cast(v4)[0] == True
    v6 = [[1.,2.,3.],[4.,5.,6.]]
    assert t.cast(v6)[0] == False

    t = FloatVarType()
    assert t.valide(v1)==False
    assert t.valide(v4)==True
    assert t.valide(v2) == True
    assert t.valide(v3) == False
    v7 = numpy.array([1,2,3], 'i')
    assert t.valide(v7) == False
    cv = t.cast(v7)
    assert t.valide(cv[1]) == True
    assert t.cast(v1)[0] == True
    v8 = [[1,2,3],[4,5,6]]
    assert t.cast(v8)[0] == False

    t = VarVectorType()
    assert t.valide(v1) == False
    assert t.valide(v2) == False
    assert t.valide (v5) == False
    assert t.valide(v6) == True
    assert t.valide([[1,2,3], [1]]) == False
    assert t.valide([[1,2,3]]) ==True
    assert t.valide(numpy.array([[1,2,3],[4,5,6]]) ) == True



##
##  Author Michel F. Sanner Jan 2009
##
import types, weakref
from Scenario2.actions import Actions
from Scenario2.keyframes import KF, Interval
from Scenario2.datatypes import DataType
from Scenario2.interpolators import Interpolator, BehaviorList

class Actor:
    """
    An Actor is an object that will modify an attribute of a Python Object
    over the course of time.

    Actions performed by the actor are represented by keyframes (representing a
    value at a given time), and intervals that can interpolate values between
    keyframes.
    
    Actors are stored in MultipleActorActions objects.

    An Actor is created with:
    - a name
    - a given Python object
    optionally:
    - an initial value
    - a datatype used to validate value for keyframes
    - default value generator function which defined how the value will modified
     by default in Intervals

     When an actor is created its Actions object is empty.

     Actors can be made active/inactive for playback.

     Actors that know how to retrieve the value from the Python object
     can be made active/inactive for recording keyframes.
     """

    _easeInOutDict = {'none':(0,0),
                      'ease in':(1,0),
                      'ease out':(0,1),
                      'ease in and out':(1,1)}


    def __init__(self, name, object, initialValue=None, datatype=None,
                 interp=None):

        self.printValuesWhenSetting = False # set to True to debug
        self.object = object
        self.name = name # actor name, has to be unique in Director.actors
        self.interp = interp
        self.initialValue = initialValue
        self.hasGetFunction = False # set to true is actor knows how to get the
                                    # value from the object
        self.recording = False # true if this actor will record keyframes
        self.playing = True  #true if the actor will set the values
        
        self._maa = None # becomes a weakref to MAA when added to MAA

        self.easeIn = False
        self.easeOut = False

        if datatype is not None:
            assert issubclass(datatype, DataType)
            self.datatype = datatype()
        else:
            self.datatype = None

        self.actions = Actions()
        self.actions.actor = weakref.ref(self)
        
        self.nbvar = None    # number of variable that are interpolated
        self.varnames = []
        self.activeVars = [] # list of booleans allowing to turn particular
                             # variables on or off
        self.interpClass = None
        args = (initialValue, initialValue)
        kw = {}
        interpolator = None
        if interp is not None:
            try:
                interpClass, args, kw = interp
                assert issubclass(interpClass, Interpolator)
                assert isinstance(args, tuple)
                assert isinstance(kw, dict)
            except TypeError:
                interpClass = interp
                assert issubclass(interpClass, Interpolator)

            interpolator = interpClass( *args, **kw) 
            self.interpClass = interpClass
            
            self.nbvar = interpClass.nbvar
            self.varnames = interpClass.varnames
            if self.nbvar:
                self.activeVars = [1]*self.nbvar

        self.behaviorList = bl = BehaviorList( *args, **{'actor':self})
        if interpolator:
            bl.addBehavior(interpolator)
            bl.configure(active = False)

        if initialValue is not None:
            kf0 = KF(0, initialValue)
            self.actions.addKeyframe(kf0)

        # FIXME .. check variable below to see if needed
        self.preStep_cb = None
        self.postStep_cb = None

        # GUI variable
        self.visible = True
        self.displayFunction = False
        self.graphHeight = 40
        self.valueFormat = None
        self.scenarioname = None

        
    def addIntervals(self, intervals, generator=None, check=True):
        """
        add a list of intervals to the actions. Intervals can be Interval
        Objects or pairs of KFs that can be specified as KF objects or
        (position, value) pairs
        """
        maxPos = 0
        actions = self.actions
        for inter in intervals:
            if isinstance(inter, Interval):
                if inter.valGen.generator is None:
                    inter.setValGenGenerator(self.behaviorList.clone())

                if inter.kf2.pos > maxPos:
                    maxPos = inter.kf2.pos
            else:
                kf1, kf2 = inter
                if not isinstance(kf1, KF):
                    kf1 = KF( *kf1 )
                if not isinstance(kf2, KF):
                    kf2 = KF( *kf2 )

                if kf2.pos > maxPos:
                    maxPos = kf2.pos

                if generator is None:
                    generator = self.behaviorList.clone()

                inter = Interval(kf1, kf2, generator=generator)

            val = actions.addInterval( inter )
            if val is False:
                return False
        
        if self._maa:
            if self._maa().getLastFrame() <= maxPos:
                self._maa().updateEndFrame( maxPos )

        return True
    

    def addKeyframe(self, kf):
        """
        None <- addKeyframe(kf)
        Adds the keyframe 'kf' to the actor's actions object
        """
        self.actions.addKeyframe(kf)
        if self._maa:
            if self._maa().getLastFrame() <= fk.pos:
                self._maa().updateEndFrame( frame )


    def addActionsAt(self, srcActions, position, check=True):
        """
        Add the actions srcActions at the given position to the actor's
        actions object. This method will add the Actor'; default value
        generator to any interval that does not have one.

        None <- actor.addActionsAt(actions, position, check=True)

        Input:
            arcAction: Actions object
            position: an integer specifying at which frame the srcAction will
                      be added
            Check: boolean

        Method used to add actions to to an actor. For every interval
        that has no value generator, this method will set it to the actor's
        default VG, and then call self.actions.addActions(actions, check) to
        add the keyframes and intervals in actions to the actor's actions.
        """
        for i in srcActions.intervals:
            if i.valGen.generator is None:
                i.setValGenGenerator(self.behaviorList.clone())
            else:
                pass # FIXME check is valgen is ok for this actor

        self.actions.addActionsAt(srcActions, position, check=check)

        
    def getLastKeyFrame(self):
        """
        KF <- actor.getLastKeyFrame()
        
        Return the last keyframe ni the actor's actions object
        """
        if len(self.actions.keyframes):
            return self.actions.keyframes[-1]
        else:
            return None

    def onAddToDirector(self):
        pass


    def setValueAt(self, frame, off=0):
        """
        None <- setValueAt(frame, off)
        argumets: frame - position in time;
                  off - offset value.
        Set the value at specified frame to the object's animated attribute.
        """
        # is called at each time step if the actor's readMode is active
        from SimPy.Simulation import now
        print 'setting ', self.name, 'for', self.name, 'at:', frame


    def setValue(self, value):
        """
        None <- setValue(value)
        set the value on the object
        """
        # is called at each time step if the actor's readMode is active
        from SimPy.Simulation import now
        print 'setting ', self.name, 'for', self.name, 'at:', now(), 'to', value


    def setEaseIn(self, val):
        """
        None <- setEaseIn(val)
        Set easeIn atribute of all value generators in the actions 
        object to the specified value (val).
        val can be 0, 1, True of False
        """ 
        assert val in [0, 1, True, False]
        self.behaviorList.easeIn = val
        for inter in self.actions.intervals:
            inter.valGen.generator.easeIn = val


    def setEaseOut(self, val):
        """
        None <- setEaseOut(val)
        Set easeOut atribute of all value generators in the actions 
        object to the specified value (val).
        val can be 0, 1, True of False
        """ 
        assert val in [0, 1, True, False]
        self.behaviorList.easeOut = val
        for inter in self.actions.intervals:
            inter.valGen.generator.easeOut = val


    def setEaseInOut(self, val):
        """
        None <- setEaseInOut(val)
        Set easeIn and easeOut atributes of all value generators in the actions 
        object.
        val can be :
           'none'             set easeIn and easeOut to 0,
           'ease in'          easeIn = True, easeOut = False,
           'ease out'         easeIn = False, easeOut = True,
           'ease in and out'  easeIn = True , easeOut = True
        """ 
        easeIn, easeOut = self._easeInOutDict[val]
        #print 'setting ease for actor', self.name, easeIn, easeOut
        self.setEaseIn(easeIn)
        self.setEaseOut(easeOut)


    def clone(self):
        """newActor <- clone()
        Return a copy of the actor (self).
        The method calls self.__class__ method. It then clones self.behaviorList and
        assignes the cloned copy to newActor.behaviorList.
        """
        if self.datatype is not None:
            dt = self.datatype.__class__
        else:
            dt = None
        newActor = self.__class__(self.name, self.object, initialValue=self.initialValue,
                                  datatype=dt, interp=self.interp)
        newActor.behaviorList = self.behaviorList.clone()
        return newActor



class CustomActor(Actor):
    """
    Custom actors are actors for which a setMethod or setFunction and a
    getFunction allowing set and get the value of the atrribute in the
    Python object that is driven by the actor can be set and gotten
    """
    def __init__(self, name, object, initialValue=None, datatype=None,
                 interp=None, setFunction=None, setMethod=None,
                 getFunction=None):
        """
Constructor of the Actor object,

arguments:
    object:            Python object on which to operate
    name:              Name of this Actor,  This name has to be unique
                       among the list of ActorDescriptor in an Director
    setFunction:       function to called at each time step.
                       The function will be called using  func(*(actor,value))
    setMethod:         method of the obect to be called at each time step.
                       The function will be called using  obj.method(value)
    getFunction=None:  [optional] function that can be called to get the
                       current value of the attribute managed by this actor
                       The function and its arguments have to be specified as a
                       3-tuple (func, *args, **kw). It will be called using
                       func(*(object,)+args), **kw) if it is a function
                       or func(*args, **kw) if it is a method
   interp              interpreter class
   initialValue        initial value of the attribute
   dataType            type of the attribute value 
   """
        
        self.getFuncTuple = None
        self.hasGetFunction = False
        if setFunction:
            assert callable(setFunction)
        self.setFunction = setFunction

        if setMethod:
            method = getattr(object, setMethod)
            assert callable(method)
        else:
            method = None
        self.setMethod = method

        if getFunction:
            self.getFuncTuple = self.checkFunction(getFunction)
            
        self.object = object
        if initialValue is None:
            if self.getFuncTuple:
                initialValue = self.getValueFromObject()

        Actor.__init__(self, name, object, datatype=datatype,
                       initialValue=initialValue, interp=interp)

        if self.getFuncTuple:
            self.hasGetFunction = True


    def clone(self):
        """newActor <- clone()
        Return a copy of the actor (self).
        The method calls self.__class__ method. It then clones self.behaviorList and
        assignes the cloned copy to newActor.behaviorList.
        """
        
        if self.setMethod is not None:
            setMethod = self.setMethod.__name__
        else:
            setMethod = None

        newActor = self.__class__(self.name, self.object, initialValue=self.initialValue,
                                  datatype=self.datatype.__class__, interp=self.interp,
                                  setFunction=self.setFunction, setMethod=setMethod,
                                  getFunction=None)
        newActor.getFuncTuple = self.getFuncTuple
        if self.getFuncTuple:
            newActor.hasGetFunction = True

        newActor.behaviorList = self.behaviorList.clone()
        return newActor


    def checkFunction(self, function):
        """
        Check if the specified 'function' is eather:
        1.callable function.
          Returned value will be a tuple: (function, (), {}).
             
        2.tuple: (function, (args), {kw}),
          where args is the function's arguments (tuple) ,
          kw - keywords dictionary.
          Returned value : (function, (args), {kw}).
        """
        # check that functionTuple is of form (func, (), {})
        try:
            f, args, kw = function
            assert callable(f)
            assert isinstance(args, tuple)
            assert isinstance(kw, dict)
        except TypeError:
            assert callable(function)
            f, args, kw = function, (), {}
        return f, args, kw

   
    def getValueFromObject(self):
        """
        value <- getValueFromObject()
        Get the current value of the animated attribute from the object
        and return it.
        """
        if not self.getFuncTuple:
            return None
        f, args, kw = self.getFuncTuple
        if type(f) == types.FunctionType or type(f) == types.BuiltinFunctionType:
            # add object as first argument to functions
            return f(*(self.object,)+args, **kw)
        elif type(f) == types.MethodType:
            return f(*args, **kw)


    def setValue(self, value):
        """
        None <- setValue(value)
        Call the self.setFunction() to set the value on the object.
        If set.selfFunction is None, this method will call self.setMethod() instead
        (if it is not None).
        """
        if self.setFunction:
            self.setFunction( *(self, value) )
        elif self.setMethod:
            self.setMethod(value)


    def setValueAt(self, frame, off):
        """
        None <- setValueAt(frame, off)
        argumets: frame - position in time;
                  off - offset value.
        The method first gets a value (at specified frame and offset) from the actions object.
        It then sets the value on the object with self.setValue() call. 
        """
        # call the function to set the value on the object
        value = self.actions.getValue(frame-off)

        if value != 'Nothing There':
            self.setValue(value)
